#
# Copyright 2019-2020 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

"""
Spectroscopy mode, combine AB or ABBA observations
"""

from astropy import wcs
from astropy.coordinates import SkyCoord
import astropy.io.fits as fits
import contextlib
import logging
import numpy as np
from scipy import ndimage
import uuid

from numina.array.wavecalib.crosscorrelation import periodic_corr1d
from numina.core import Result
from numina.core import Parameter
import numina.array.combine as combine
from numina.array.distortion import shift_image2d

import emirdrp.requirements as reqs
from emirdrp.core.recipe import EmirRecipe
import emirdrp.datamodel
import emirdrp.products as prods
from numina.processing.combine import combine_imgs
from emirdrp.processing.wavecal.apply_rectwv_coeff import apply_rectwv_coeff
from emirdrp.processing.wavecal.median_slitlets_rectified import \
    median_slitlets_rectified
from emirdrp.processing.wavecal.rectwv_coeff_to_ds9 import save_four_ds9
from emirdrp.processing.wavecal.rectwv_coeff_to_ds9 import \
    save_spectral_lines_ds9
from emirdrp.processing.wavecal.useful_mos_xpixels import useful_mos_xpixels

from emirdrp.core import EMIR_NAXIS1
from emirdrp.core import EMIR_NAXIS2


def get_isky(i, pattern):
    """Return index number of image to be employed as sky"""

    if pattern == 'AB':
        if (i + 1) % 2 == 0:
            return i - 1
        else:
            return i + 1
    elif pattern == 'ABBA':
        n_previous_sequences = (i + 1) // 4
        if (i + 1) % 4 == 0:
            n_previous_sequences -= 1
        ieff = i - n_previous_sequences * 4
        if ieff == 0:
            iout = i + 1
        elif ieff == 1:
            iout = i - 1
        elif ieff == 2:
            iout = i + 1
        elif ieff == 3:
            iout = i - 1
        else:
            raise ValueError('Unexpected ieff value={}'.format(ieff))
    else:
        raise ValueError('Unexpected pattern: {}'.format(pattern))

    return iout


def compute_wcs_offsets(hduls):
    """Compute offsets and spatial scales from WCS info in image headers.

    Compute offset (arcsec) between the first image in the sequence and
    the rest of the images, as well as the spatial scale (arcsec/pix)
    within each individual image.

    """

    nimages = len(hduls)
    c0 = None
    list_sep_arcsec = []
    list_spatial_scales_arcsecperpix = []
    for i in range(nimages):
        wcsh = wcs.WCS(hduls[i][0].header)
        ra, dec = wcsh.wcs_pix2world(EMIR_NAXIS1 / 2 + 0.5,
                                     EMIR_NAXIS2 / 2 + 0.5, 1)
        ra_, dec_ = wcsh.wcs_pix2world(EMIR_NAXIS1 / 2 + 0.5,
                                       EMIR_NAXIS2 / 2 + 1.5, 1)
        if i == 0:
            c0 = SkyCoord(ra, dec, unit="deg")
        ci = SkyCoord(ra, dec, unit="deg")
        ci_ = SkyCoord(ra_, dec_, unit="deg")
        sep = ci.separation(c0).arcsec
        scale = ci.separation(ci_).arcsec
        list_sep_arcsec.append(round(sep, 4))
        list_spatial_scales_arcsecperpix.append(round(scale, 7))

    sep_arcsec = np.array(list_sep_arcsec)
    spatial_scales_arcsecperpix = np.array(list_spatial_scales_arcsecperpix)

    return sep_arcsec, spatial_scales_arcsecperpix


class ABBASpectraRectwv(EmirRecipe):
    """Process images in AB or ABBA  mode applying wavelength calibration

    Note that in this case the wavelength calibration has already been
    determined.

    """

    logger = logging.getLogger(__name__)

    obresult = reqs.ObservationResultRequirement()
    master_bpm = reqs.MasterBadPixelMaskRequirement()
    master_bias = reqs.MasterBiasRequirement()
    master_flat = reqs.MasterSpectralFlatFieldRequirement()
    rectwv_coeff = reqs.RectWaveCoeffRequirement(optional=True)
    list_rectwv_coeff = reqs.ListOfRectWaveCoeffRequirement(optional=True)

    pattern = Parameter(
        'ABBA',
        description='Observation pattern',
        choices=['AB', 'ABBA']
    )
    # do not allow 'sum' as combination method in order to avoid
    # confusion with number of counts in resulting image
    method = Parameter(
        'mean',
        description='Combination method',
        choices=['mean', 'median', 'sigmaclip']
    )
    method_kwargs = Parameter(
        dict(),
        description='Arguments for combination method',
        optional=True
    )
    refine_target_along_slitlet = Parameter(
        dict(),
        description='Parameters to refine location of target along the slitlet'
    )

    reduced_mos_abba = Result(prods.ProcessedMOS)
    reduced_mos_abba_combined = Result(prods.ProcessedMOS)

    def run(self, rinput):

        nimages = len(rinput.obresult.frames)
        pattern = rinput.pattern
        pattern_length = len(pattern)

        # check combination method
        if rinput.method != 'sigmaclip':
            if rinput.method_kwargs != {}:
                raise ValueError('Unexpected method_kwargs={}'.format(
                    rinput.method_kwargs))

        # check pattern sequence matches number of images
        if nimages % pattern_length != 0:
            raise ValueError('Number of images is not a multiple of pattern '
                             'length: {}, {}'.format(nimages, pattern_length))
        nsequences = nimages // pattern_length

        # check rectification and wavelength calibration information
        if rinput.rectwv_coeff is None and rinput.list_rectwv_coeff is None:
            raise ValueError('No rectwv_coeff nor list_rectwv_coeff data have '
                             'been provided')
        elif rinput.rectwv_coeff is not None and \
                rinput.list_rectwv_coeff is not None:
            raise ValueError("rectwv_coeff and list_rectwv_coeff cannot be "
                             "used simultaneously")
        elif rinput.rectwv_coeff is not None:
            list_rectwv_coeff = [rinput.rectwv_coeff] * nimages
        elif rinput.list_rectwv_coeff is not None:
            if len(rinput.list_rectwv_coeff) != nimages:
                raise ValueError("Unexpected number of rectwv_coeff files "
                                 "in list_rectwv_coeff")
            else:
                list_rectwv_coeff = rinput.list_rectwv_coeff
                # check filter and grism are the same in all JSON files
                for item in ['grism', 'filter']:
                    list_values = []
                    for calib in list_rectwv_coeff:
                        list_values.append(calib.tags[item])
                    if len(set(list_values)) != 1:
                        raise ValueError(
                            'list_rectwv_coeff contains coefficients for '
                            'different {}s'.format(item)
                        )
        else:
            raise ValueError("Unexpected error!")

        # grism and filter names
        grism_name = list_rectwv_coeff[0].tags['grism']
        filter_name = list_rectwv_coeff[0].tags['filter']

        # compute offsets from WCS info in image headers
        with contextlib.ExitStack() as stack:
            hduls = [stack.enter_context(fname.open()) for fname in
                     rinput.obresult.frames]
            sep_arcsec, spatial_scales = compute_wcs_offsets(hduls)

        sep_pixel = np.round(sep_arcsec / spatial_scales, 6)

        for i in range(nimages):
            self.logger.info(list_rectwv_coeff[i])
        self.logger.info('observation pattern..................: {}'.format(
            pattern))
        self.logger.info('nsequences...........................: {}'.format(
            nsequences))
        self.logger.info('offsets (arcsec, from WCS)...........: {}'.format(
            sep_arcsec))
        self.logger.info('spatial scales (arcsec/pix, from WCS): {}'.format(
            spatial_scales))
        self.logger.info('spatial scales (pixels, from WCS)....: {}'.format(
            sep_pixel))

        full_set = pattern * nsequences
        self.logger.info('full set of images...................: {}'.format(
            full_set))

        # basic parameters to determine useful pixels in the wavelength
        # direction
        dict_rtas = rinput.refine_target_along_slitlet

        valid_keys = [
            'npix_removed_near_ohlines',
            'nwidth_medfilt',
            'save_individual_images',
            'ab_different_target',
            'vpix_region_a_target',
            'vpix_region_a_sky',
            'vpix_region_b_target',
            'vpix_region_b_sky',
            'list_valid_wvregions_a',
            'list_valid_wvregions_b'
        ]
        for dumkey in dict_rtas.keys():
            if dumkey not in valid_keys:
                raise ValueError('Unexpected key={}'.format(dumkey))

        if 'vpix_region_a_target' in dict_rtas.keys():
            vpix_region_a_target = dict_rtas['vpix_region_a_target']
        else:
            vpix_region_a_target = None

        if 'vpix_region_a_sky' in dict_rtas.keys():
            vpix_region_a_sky = dict_rtas['vpix_region_a_sky']
        else:
            vpix_region_a_sky = None

        if 'vpix_region_b_target' in dict_rtas.keys():
            vpix_region_b_target = dict_rtas['vpix_region_b_target']
        else:
            vpix_region_b_target = None

        if 'vpix_region_b_sky' in dict_rtas.keys():
            vpix_region_b_sky = dict_rtas['vpix_region_b_sky']
        else:
            vpix_region_b_sky = None

        if 'ab_different_target' in dict_rtas.keys():
            ab_different_target = int(dict_rtas['ab_different_target'])
            if ab_different_target not in [-1, 0, 1, 9]:
                raise ValueError('Invalid ab_different_target={} value'.format(
                    ab_different_target))
        else:
            raise ValueError('Missing ab_different_target value')

        try:
            npix_removed_near_ohlines = \
                int(dict_rtas['npix_removed_near_ohlines'])
        except KeyError:
            npix_removed_near_ohlines = 0
        except ValueError:
            raise ValueError(
                'wrong value: npix_removed_near_ohlines='
                '{}'.format(dict_rtas['npix_removed_near_ohlines'])
            )

        if 'list_valid_wvregions_a' in dict_rtas.keys():
            if vpix_region_a_target is None:
                raise ValueError('Unexpected list_valid_wvregions_a when '
                                 'vpix_region_a_target is not set')
            list_valid_wvregions_a = dict_rtas['list_valid_wvregions_a']
        else:
            list_valid_wvregions_a = None

        if 'list_valid_wvregions_b' in dict_rtas.keys():
            if vpix_region_b_target is None:
                raise ValueError('Unexpected list_valid_wvregions_b when '
                                 'vpix_region_b_target is not set')

            list_valid_wvregions_b = dict_rtas['list_valid_wvregions_b']
        else:
            list_valid_wvregions_b = None

        try:
            nwidth_medfilt = int(dict_rtas['nwidth_medfilt'])
        except KeyError:
            nwidth_medfilt = 0
        except ValueError:
            raise ValueError('wrong value: nwidth_medfilt={}'.format(
                dict_rtas['nwidth_medfilt']))

        try:
            save_individual_images = int(dict_rtas['save_individual_images'])
        except KeyError:
            save_individual_images = 0
        except ValueError:
            raise ValueError('wrong value: save_individual_images={}'.format(
                dict_rtas['save_individual_images']))

        self.logger.info('npix_removed_near_ohlines: {}'.format(
            npix_removed_near_ohlines))
        self.logger.info('nwidth_medfilt: {}'.format(nwidth_medfilt))
        self.logger.info('vpix_region_a_target: {}'.format(
            vpix_region_a_target))
        self.logger.info('vpix_region_b_target: {}'.format(
            vpix_region_b_target))
        self.logger.info('list_valid_wvregions_a: {}'.format(
            list_valid_wvregions_a))
        self.logger.info('list_valid_wvregions_b: {}'.format(
            list_valid_wvregions_b))

        # build object to proceed with bpm, bias, dark and flat
        flow = self.init_filters(rinput)

        # basic reduction, rectification and wavelength calibration of
        # all the individual images
        list_reduced_mos_images = []
        self.logger.info('starting reduction of individual images')
        for i, char in enumerate(full_set):
            frame = rinput.obresult.frames[i]
            self.logger.info('image {} ({} of {})'.format(char, i + 1,
                                                          nimages))
            self.logger.info('image: {}'.format(frame.filename))
            with frame.open() as f:
                newimg = fits.HDUList([ext.copy() for ext in f])
            base_header = newimg[0].header
            grism_name_ = base_header['grism']
            if grism_name_ != grism_name:
                raise ValueError('Incompatible grism name in '
                                 'rectwv_coeff.json file and FITS image')
            filter_name_ = base_header['filter']
            if filter_name_ != filter_name:
                raise ValueError('Incompatible filter name in '
                                 'rectwv_coeff.json file and FITS image')
            hdu = newimg[0]
            hdu.header['UUID'] = str(uuid.uuid1())
            # basic reduction
            reduced_image = flow(newimg)
            hdr = reduced_image[0].header
            self.set_base_headers(hdr)
            # rectification and wavelength calibration
            reduced_mos_image = apply_rectwv_coeff(
                reduced_image,
                list_rectwv_coeff[i]
            )
            if save_individual_images != 0:
                self.save_intermediate_img(
                    reduced_mos_image,
                    'reduced_mos_image_' + char + '_' +
                    frame.filename[:10] + '.fits'
                )
            list_reduced_mos_images.append(reduced_mos_image)

        # intermediate PDF file with crosscorrelation plots
        if self.intermediate_results:
            from matplotlib.backends.backend_pdf import PdfPages
            pdf = PdfPages('crosscorrelation_ab.pdf')
        else:
            pdf = None

        # compute offsets between images
        self.logger.info('computing offsets between individual images')
        first_a = True
        first_b = True
        xisok_a = None
        xisok_b = None
        reference_profile_a = None
        reference_profile_b = None
        refine_a = vpix_region_a_target is not None
        refine_b = vpix_region_b_target is not None
        list_offsets = []
        for i, char in enumerate(full_set):
            if char == 'A' and refine_a:
                refine_image = True
            elif char == 'B' and refine_b:
                refine_image = True
            else:
                refine_image = False
            if refine_image:
                frame = rinput.obresult.frames[i]
                isky = get_isky(i, pattern)
                frame_sky = rinput.obresult.frames[isky]
                self.logger.info('image {} ({} of {})'.format(char, i + 1,
                                                              nimages))
                self.logger.info('image: {}'.format(frame.filename))
                self.logger.info('(sky): {}'.format(frame_sky.filename))
                data = list_reduced_mos_images[i][0].data.copy()
                data_sky = list_reduced_mos_images[isky][0].data
                data -= data_sky
                base_header = list_reduced_mos_images[i][0].header

                # get useful pixels in the wavelength direction
                if char == 'A' and first_a:
                    xisok_a = useful_mos_xpixels(
                        data,
                        base_header,
                        vpix_region=vpix_region_a_target,
                        npix_removed_near_ohlines=npix_removed_near_ohlines,
                        list_valid_wvregions=list_valid_wvregions_a,
                        debugplot=0
                    )
                elif char == 'B' and first_b:
                    xisok_b = useful_mos_xpixels(
                        data,
                        base_header,
                        vpix_region=vpix_region_b_target,
                        npix_removed_near_ohlines=npix_removed_near_ohlines,
                        list_valid_wvregions=list_valid_wvregions_b,
                        debugplot=0
                    )

                if char == 'A':
                    nsmin = vpix_region_a_target[0]
                    nsmax = vpix_region_a_target[1]
                    xisok = xisok_a
                    if vpix_region_a_sky is not None:
                        nsmin_sky = vpix_region_a_sky[0]
                        nsmax_sky = vpix_region_a_sky[1]
                        skysubtraction = True
                    else:
                        nsmin_sky = None
                        nsmax_sky = None
                        skysubtraction = False
                elif char == 'B':
                    nsmin = vpix_region_b_target[0]
                    nsmax = vpix_region_b_target[1]
                    xisok = xisok_b
                    if vpix_region_b_sky is not None:
                        nsmin_sky = vpix_region_b_sky[0]
                        nsmax_sky = vpix_region_b_sky[1]
                        skysubtraction = True
                    else:
                        nsmin_sky = None
                        nsmax_sky = None
                        skysubtraction = False
                else:
                    raise ValueError('Unexpected char value: {}'.format(char))

                # initial slitlet region
                slitlet2d = data[(nsmin - 1):nsmax, :].copy()
                if skysubtraction:
                    slitlet2d_sky = data[(nsmin_sky - 1):nsmax_sky, :].copy()
                    median_sky = np.median(slitlet2d_sky, axis=0)
                    slitlet2d -= median_sky

                # selected wavelength regions after blocking OH lines
                slitlet2d_blocked = slitlet2d[:, xisok]

                # apply median filter in the X direction
                if nwidth_medfilt > 1:
                    slitlet2d_blocked_smoothed = ndimage.filters.median_filter(
                        slitlet2d_blocked, size=(1, nwidth_medfilt)
                    )
                else:
                    slitlet2d_blocked_smoothed = slitlet2d_blocked

                # ---
                # convert to 1D series of spatial profiles (note: this
                # option does not work very well when the signal is low;
                # a simple mean profile does a better job)
                # profile = slitlet2d_blocked_smoothed.transpose().ravel()
                # ---
                profile = np.mean(slitlet2d_blocked_smoothed, axis=1)
                if char == 'A' and first_a:
                    reference_profile_a = profile.copy()
                elif char == 'B' and first_b:
                    reference_profile_b = profile.copy()
                # ---
                # # To write pickle file
                # import pickle
                # with open(frame.filename[:10] + '_profile.pkl', 'wb') as f:
                #     pickle.dump(profile, f)
                # # To read pickle file
                # with open('test.pkl', 'rb') as f:
                #     x = pickle.load(f)
                # ---

                # crosscorrelation to find offset
                naround_zero = (nsmax - nsmin) // 3
                if char == 'A':
                    reference_profile = reference_profile_a
                elif char == 'B':
                    reference_profile = reference_profile_b
                else:
                    raise ValueError('Unexpected char value: {}'.format(char))
                offset, fpeak = periodic_corr1d(
                    sp_reference=reference_profile,
                    sp_offset=profile,
                    remove_mean=False,
                    frac_cosbell=0.10,
                    zero_padding=11,
                    fminmax=None,
                    nfit_peak=5,
                    naround_zero=naround_zero,
                    sp_label='spatial profile',
                    plottitle='Image #{} (type {}), {}'.format(
                        i + 1, char, frame.filename[:10]),
                    pdf=pdf
                )
                # round to 4 decimal places
                if abs(offset) < 1E-4:
                    offset = 0.0   # avoid -0.0
                else:
                    offset = round(offset, 4)
                list_offsets.append(offset)

                # end of loop
                if char == 'A' and first_a:
                    first_a = False
                elif char == 'B' and first_b:
                    first_b = False
            else:
                list_offsets.append(0.0)
        self.logger.info('computed offsets: {}'.format(list_offsets))

        self.logger.info('correcting vertical offsets between individual '
                         'images')
        list_a = []
        list_b = []
        for i, (char, offset) in enumerate(zip(full_set, list_offsets)):
            frame = rinput.obresult.frames[i]
            self.logger.info('image {} ({} of {})'.format(char, i + 1,
                                                          nimages))
            self.logger.info('image: {}'.format(frame.filename))
            reduced_mos_image = list_reduced_mos_images[i]
            data = reduced_mos_image[0].data
            base_header = reduced_mos_image[0].header
            self.logger.info(
                'correcting vertical offset (pixesl): {}'.format(offset))
            if offset != 0:
                reduced_mos_image[0].data = shift_image2d(
                    data,
                    yoffset=-offset
                ).astype('float32')
            base_header['HISTORY'] = 'Applying voffset_pix {}'.format(offset)
            if save_individual_images != 0:
                self.save_intermediate_img(
                    reduced_mos_image,
                    'reduced_mos_image_refined_' + char + '_' +
                    frame.filename[:10] + '.fits'
                )

            # store reduced_mos_image
            if char == 'A':
                list_a.append(reduced_mos_image)
            elif char == 'B':
                list_b.append(reduced_mos_image)
            else:
                raise ValueError('Unexpected char value: {}'.format(char))

        # combination method
        method = getattr(combine, rinput.method)
        method_kwargs = rinput.method_kwargs

        # final combination of A images
        self.logger.info('combining individual A images')
        reduced_mos_image_a = combine_imgs(
            list_a,
            method=method,
            method_kwargs=method_kwargs,
            errors=False,
            prolog=None
        )
        self.save_intermediate_img(reduced_mos_image_a,
                                   'reduced_mos_image_a.fits')

        # final combination of B images
        self.logger.info('combining individual B images')
        reduced_mos_image_b = combine_imgs(
            list_b,
            method=method,
            method_kwargs=method_kwargs,
            errors=False,
            prolog=None
        )
        self.save_intermediate_img(reduced_mos_image_b,
                                   'reduced_mos_image_b.fits')

        self.logger.info('mixing A and B spectra')
        header_a = reduced_mos_image_a[0].header
        header_b = reduced_mos_image_b[0].header
        data_a = reduced_mos_image_a[0].data.astype('float32')
        data_b = reduced_mos_image_b[0].data.astype('float32')

        reduced_mos_abba_data = data_a - data_b

        # update reduced mos image header
        reduced_mos_abba = self.create_mos_abba_image(
            rinput,
            dict_rtas,
            reduced_mos_abba_data,
            header_a, header_b,
            pattern,
            full_set,
            list_offsets,
            voffset_pix=0
        )

        # combine A and B data by shifting B on top of A
        if abs(ab_different_target) == 0:
            len_prof_a = len(reference_profile_a)
            len_prof_b = len(reference_profile_b)
            if len_prof_a == len_prof_b:
                reference_profile = reference_profile_a
                profile = reference_profile_b
                naround_zero = len_prof_a // 3
            elif len_prof_a > len_prof_b:
                ndiff = len_prof_a - len_prof_b
                reference_profile = reference_profile_a
                profile = np.concatenate(
                    (reference_profile_b, np.zeros(ndiff, dtype='float'))
                )
                naround_zero = len_prof_a // 3
            else:
                ndiff = len_prof_b - len_prof_a
                reference_profile = np.concatenate(
                    (reference_profile_a, np.zeros(ndiff, dtype='float'))
                )
                profile = reference_profile_b
                naround_zero = len_prof_b // 3
            offset, fpeak = periodic_corr1d(
                sp_reference=reference_profile,
                sp_offset=profile,
                remove_mean=False,
                frac_cosbell=0.10,
                zero_padding=11,
                fminmax=None,
                nfit_peak=5,
                naround_zero=naround_zero,
                sp_label='spatial profile',
                plottitle='Comparison of A and B profiles',
                pdf=pdf
            )
            voffset_pix = vpix_region_b_target[0] - vpix_region_a_target[0]
            voffset_pix += offset
        elif abs(ab_different_target) == 1:
            # apply nominal offset (from WCS info) between first A
            # and first B
            voffset_pix = sep_pixel[1] * ab_different_target
        elif ab_different_target == 9:
            voffset_pix = None
        else:
            raise ValueError('Invalid ab_different_target={}'.format(
                ab_different_target))

        # close output PDF file
        if pdf is not None:
            pdf.close()

        if voffset_pix is not None:
            self.logger.info(
                'correcting vertical offset (pixesl): {}'.format(voffset_pix))
            shifted_a_minus_b_data = shift_image2d(
                reduced_mos_abba_data,
                yoffset=-voffset_pix,
            ).astype('float32')
            reduced_mos_abba_combined_data = \
                reduced_mos_abba_data - shifted_a_minus_b_data
            # scale signal to exposure of a single image
            reduced_mos_abba_combined_data /= 2.0
        else:
            reduced_mos_abba_combined_data = None

        # update reduced mos combined image header
        reduced_mos_abba_combined = self.create_mos_abba_image(
            rinput,
            dict_rtas,
            reduced_mos_abba_combined_data,
            header_a, header_b,
            pattern,
            full_set,
            list_offsets,
            voffset_pix
        )

        # ds9 region files (to be saved in the work directory)
        if self.intermediate_results:
            save_four_ds9(list_rectwv_coeff[0])
            save_spectral_lines_ds9(list_rectwv_coeff[0])

        # compute median spectra employing the useful region of the
        # rectified image
        if self.intermediate_results:
            for imode, outfile in enumerate(['median_spectra_full',
                                             'median_spectra_slitlets',
                                             'median_spectrum_slitlets']):
                median_image = median_slitlets_rectified(
                    reduced_mos_abba, mode=imode
                )
                self.save_intermediate_img(median_image, outfile + '.fits')

        # save results in results directory
        self.logger.info('end rect.+wavecal. reduction of ABBA spectra')
        result = self.create_result(
            reduced_mos_abba=reduced_mos_abba,
            reduced_mos_abba_combined=reduced_mos_abba_combined
        )
        return result

    def create_mos_abba_image(self, rinput, dict_rtas, reduced_mos_abba_data,
                              header_a, header_b,
                              pattern, full_set, list_offsets, voffset_pix):
        with contextlib.ExitStack() as stack:
            hduls = [stack.enter_context(fname.open()) for fname in
                     rinput.obresult.frames]
            # Copy the first image
            result_img = fits.HDUList([ext.copy() for ext in hduls[0]])
            hdu = result_img[0]
            hdu.data = reduced_mos_abba_data
            self.set_base_headers(hdu.header)

            # check consistency of wavelength calibration paramenters
            for param in ['crpix1', 'crval1', 'cdelt1']:
                if header_a[param] != header_b[param]:
                    raise ValueError('Headers of A and B images have different '
                                     'values of {}'.format(param))
            self.logger.debug('update result header')
            crpix1 = header_a['crpix1']
            crval1 = header_a['crval1']
            cdelt1 = header_a['cdelt1']

            # update wavelength calibration in FITS header
            for keyword in ['crval1', 'crpix1', 'crval2', 'crpix2']:
                if keyword in hdu.header:
                    hdu.header.remove(keyword)
            hdu.header['crpix1'] = (crpix1, 'reference pixel')
            hdu.header['crval1'] = (crval1, 'central wavelength at crpix1')
            hdu.header['cdelt1'] = (cdelt1, 'linear dispersion (Angstrom/pixel)')
            hdu.header['cunit1'] = ('Angstrom', 'units along axis1')
            hdu.header['ctype1'] = 'WAVE'
            hdu.header['crpix2'] = (0.0, 'reference pixel')
            hdu.header['crval2'] = (0.0, 'central value at crpix2')
            hdu.header['cdelt2'] = (1.0, 'increment')
            hdu.header['ctype2'] = 'PIXEL'
            hdu.header['cunit2'] = ('Pixel', 'units along axis2')
            for keyword in ['cd1_1', 'cd1_2', 'cd2_1', 'cd2_2',
                            'PCD1_1', 'PCD1_2', 'PCD2_1', 'PCD2_2',
                            'PCRPIX1', 'PCRPIX2']:
                if keyword in hdu.header:
                    hdu.header.remove(keyword)

            # update additional keywords
            hdu.header['UUID'] = str(uuid.uuid1())
            hdu.header['OBSMODE'] = pattern + ' pattern'
            hdu.header['TSUTC2'] = hduls[-1][0].header['TSUTC2']
            hdu.header['NUM-NCOM'] = (len(hduls), 'Number of combined frames')

            # update history
            hdu.header['HISTORY'] = "Processed " + pattern + " pattern"
            hdu.header['HISTORY'] = '--- Reduction of A images ---'
            for line in header_a['HISTORY']:
                hdu.header['HISTORY'] = line
            hdu.header['HISTORY'] = '--- Reduction of B images ---'
            for line in header_b['HISTORY']:
                hdu.header['HISTORY'] = line
            hdu.header['HISTORY'] = '--- Combination of ABBA images ---'
            for key in dict_rtas:
                hdu.header['HISTORY'] = '{}: {}'.format(key, dict_rtas[key])
            dm = emirdrp.datamodel.EmirDataModel()
            for img, key, offset in zip(hduls, full_set, list_offsets):
                imgid = dm.get_imgid(img)
                hdu.header['HISTORY'] = \
                    "Image '{}' is '{}', with voffset_pix {}".format(
                        imgid, key, offset)

        if voffset_pix is not None and voffset_pix != 0:
            hdu.header['HISTORY'] = '--- Combination of AB spectra ---'
            hdu.header['HISTORY'] = "voffset_pix between A and B {}".format(
                voffset_pix)
        hdu.header.add_history('--- rinput.stored() (BEGIN) ---')
        for item in rinput.stored():
            value = getattr(rinput, item)
            cline = '{}: {}'.format(item, value)
            hdu.header.add_history(cline)
        hdu.header.add_history('--- rinput.stored() (END) ---')
        return result_img

    def set_base_headers(self, hdr):
        newhdr = super(ABBASpectraRectwv, self).set_base_headers(hdr)
        # Update EXP to 0
        newhdr['EXP'] = 0
        return newhdr

    def extract_tags_from_obsres(self, obsres, tag_keys):
        # this function is necessary to make use of recipe requirements
        # that are provided as lists
        final_tags = []
        for frame in obsres.frames:
            ref_img = frame.open()
            tag = self.extract_tags_from_ref(
                ref_img,
                tag_keys,
                base=obsres.labels
            )
            final_tags.append(tag)
        return final_tags


class ABBASpectraFastRectwv(EmirRecipe):
    """Process AB or ABBA images applying a single wavelength calibration

    Note that the wavelength calibration has already been determined.

    """

    logger = logging.getLogger(__name__)

    obresult = reqs.ObservationResultRequirement()
    master_bpm = reqs.MasterBadPixelMaskRequirement()
    master_bias = reqs.MasterBiasRequirement()
    master_flat = reqs.MasterSpectralFlatFieldRequirement()
    rectwv_coeff = reqs.RectWaveCoeffRequirement()

    pattern = Parameter(
        'ABBA',
        description='Observation pattern',
        choices=['AB', 'ABBA']
    )
    # do not allow 'sum' as combination method in order to avoid
    # confusion with number of counts in resulting image
    method = Parameter(
        'mean',
        description='Combination method',
        choices=['mean', 'median', 'sigmaclip']
    )
    method_kwargs = Parameter(
        dict(),
        description='Arguments for combination method'
    )
    voffset_pix = Parameter(
        0.0,
        description='Shift (pixels) to move A into B',
        optional=True
    )

    reduced_mos_abba = Result(prods.ProcessedMOS)
    reduced_mos_abba_combined = Result(prods.ProcessedMOS)

    def run(self, rinput):

        nimages = len(rinput.obresult.frames)
        pattern = rinput.pattern
        pattern_length = len(pattern)

        # check combination method
        if rinput.method != 'sigmaclip':
            if rinput.method_kwargs != {}:
                raise ValueError('Unexpected method_kwargs={}'.format(
                    rinput.method_kwargs))

        # check pattern sequence matches number of images
        if nimages % pattern_length != 0:
            raise ValueError('Number of images is not a multiple of pattern '
                             'length: {}, {}'.format(nimages, pattern_length))
        nsequences = nimages // pattern_length

        rectwv_coeff = rinput.rectwv_coeff
        self.logger.info(rectwv_coeff)
        self.logger.info('observation pattern: {}'.format(pattern))
        self.logger.info('nsequences.........: {}'.format(nsequences))

        full_set = pattern * nsequences
        self.logger.info('full set of images.: {}'.format(full_set))

        # build object to proceed with bpm, bias, dark and flat
        flow = self.init_filters(rinput)

        # available combination methods
        method = getattr(combine, rinput.method)
        method_kwargs = rinput.method_kwargs

        # basic reduction of A images
        list_a = [rinput.obresult.frames[i] for i, char in enumerate(full_set)
                  if char == 'A']
        with contextlib.ExitStack() as stack:
            self.logger.info('starting basic reduction of A images')
            hduls = [stack.enter_context(fname.open()) for fname in list_a]
            reduced_image_a = combine_imgs(
                hduls,
                method=method,
                method_kwargs=method_kwargs,
                errors=False,
                prolog=None
            )
        reduced_image_a = flow(reduced_image_a)
        hdr = reduced_image_a[0].header
        self.set_base_headers(hdr)

        # basic reduction of B images
        list_b = [rinput.obresult.frames[i] for i, char in
                  enumerate(full_set) if char == 'B']
        with contextlib.ExitStack() as stack:
            self.logger.info('starting basic reduction of B images')
            hduls = [stack.enter_context(fname.open()) for fname in list_b]
            reduced_image_b = combine_imgs(
                hduls,
                method=method,
                method_kwargs=method_kwargs,
                errors=False,
                prolog=None
            )
        reduced_image_b = flow(reduced_image_b)
        hdr = reduced_image_b[0].header
        self.set_base_headers(hdr)

        # save intermediate reduced_image_a and reduced_image_b
        self.save_intermediate_img(reduced_image_a, 'reduced_image_a.fits')
        self.save_intermediate_img(reduced_image_b, 'reduced_image_b.fits')

        # computation of A-B
        header_a = reduced_image_a[0].header
        header_b = reduced_image_b[0].header
        data_a = reduced_image_a[0].data.astype('float32')
        data_b = reduced_image_b[0].data.astype('float32')
        reduced_data = data_a - data_b

        # update reduced image header
        reduced_image = self.create_reduced_image(
            rinput,
            reduced_data,
            header_a,
            header_b,
            rinput.pattern,
            full_set,
            voffset_pix=0,
            header_mos_abba=None,
        )

        # save intermediate image in work directory
        self.save_intermediate_img(reduced_image, 'reduced_image.fits')

        # apply rectification and wavelength calibration
        self.logger.info('begin rect.+wavecal. reduction of ABBA spectra')
        reduced_mos_abba = apply_rectwv_coeff(
            reduced_image,
            rectwv_coeff
        )
        header_mos_abba = reduced_mos_abba[0].header

        # combine A and B data by shifting B on top of A
        voffset_pix = rinput.voffset_pix

        if voffset_pix is not None and voffset_pix != 0:
            self.logger.info(
                'correcting vertical offset (pixesl): {}'.format(voffset_pix))
            reduced_mos_abba_data = reduced_mos_abba[0].data.astype('float32')
            shifted_a_minus_b_data = shift_image2d(
                reduced_mos_abba_data,
                yoffset=-voffset_pix,
            ).astype('float32')
            reduced_mos_abba_combined_data = \
                reduced_mos_abba_data - shifted_a_minus_b_data
            # scale signal to exposure of a single image
            reduced_mos_abba_combined_data /= 2.0
        else:
            reduced_mos_abba_combined_data = None

        # update reduced combined image header
        reduced_mos_abba_combined = self.create_reduced_image(
            rinput,
            reduced_mos_abba_combined_data,
            header_a,
            header_b,
            rinput.pattern,
            full_set,
            voffset_pix,
            header_mos_abba=header_mos_abba
        )

        # ds9 region files (to be saved in the work directory)
        if self.intermediate_results:
            save_four_ds9(rectwv_coeff)
            save_spectral_lines_ds9(rectwv_coeff)

        # compute median spectra employing the useful region of the
        # rectified image
        if self.intermediate_results:
            for imode, outfile in enumerate(['median_spectra_full',
                                             'median_spectra_slitlets',
                                             'median_spectrum_slitlets']):
                median_image = median_slitlets_rectified(
                    reduced_mos_abba, mode=imode
                )
                self.save_intermediate_img(median_image, outfile + '.fits')

        # save results in results directory
        self.logger.info('end rect.+wavecal. reduction of ABBA spectra')
        result = self.create_result(
            reduced_mos_abba=reduced_mos_abba,
            reduced_mos_abba_combined=reduced_mos_abba_combined
        )
        return result

    def create_reduced_image(self, rinput, reduced_data,
                             header_a, header_b,
                             pattern, full_set,
                             voffset_pix, header_mos_abba):
        with contextlib.ExitStack() as stack:
            hduls = [stack.enter_context(fname.open()) for fname in
                     rinput.obresult.frames]

            # Copy the first image
            result_img = fits.HDUList([ext.copy() for ext in hduls[0]])
            hdu = result_img[0]
            hdu.data = reduced_data
            self.set_base_headers(hdu.header)

            self.logger.debug('update result header')
            if header_mos_abba is not None:
                self.logger.debug('update result header')
                crpix1 = header_mos_abba['crpix1']
                crval1 = header_mos_abba['crval1']
                cdelt1 = header_mos_abba['cdelt1']

                # update wavelength calibration in FITS header
                for keyword in ['crval1', 'crpix1', 'crval2', 'crpix2']:
                    if keyword in hdu.header:
                        hdu.header.remove(keyword)
                hdu.header['crpix1'] = (crpix1, 'reference pixel')
                hdu.header['crval1'] = (crval1, 'central wavelength at crpix1')
                hdu.header['cdelt1'] = \
                    (cdelt1, 'linear dispersion (Angstrom/pixel)')
                hdu.header['cunit1'] = ('Angstrom', 'units along axis1')
                hdu.header['ctype1'] = 'WAVE'
                hdu.header['crpix2'] = (0.0, 'reference pixel')
                hdu.header['crval2'] = (0.0, 'central value at crpix2')
                hdu.header['cdelt2'] = (1.0, 'increment')
                hdu.header['ctype2'] = 'PIXEL'
                hdu.header['cunit2'] = ('Pixel', 'units along axis2')
                for keyword in ['cd1_1', 'cd1_2', 'cd2_1', 'cd2_2',
                                'PCD1_1', 'PCD1_2', 'PCD2_1', 'PCD2_2',
                                'PCRPIX1', 'PCRPIX2']:
                    if keyword in hdu.header:
                        hdu.header.remove(keyword)

            # update additional keywords
            hdu.header['UUID'] = str(uuid.uuid1())
            hdu.header['OBSMODE'] = pattern + ' pattern'
            hdu.header['TSUTC2'] = hduls[-1][0].header['TSUTC2']
            hdu.header['history'] = "Processed " + pattern + " pattern"
            hdu.header['NUM-NCOM'] = (len(hduls), 'Number of combined frames')

            # update history
            dm = emirdrp.datamodel.EmirDataModel()
            for img, key in zip(hduls, full_set):
                imgid = dm.get_imgid(img)
                hdu.header['HISTORY'] = "Image '{}' is '{}'".format(imgid, key)

        hdu.header['HISTORY'] = "Processed " + pattern + " pattern"
        hdu.header['HISTORY'] = '--- Reduction of A images ---'
        for line in header_a['HISTORY']:
            hdu.header['HISTORY'] = line
        hdu.header['HISTORY'] = '--- Reduction of B images ---'
        for line in header_b['HISTORY']:
            hdu.header['HISTORY'] = line
        if voffset_pix is not None and voffset_pix != 0:
            hdu.header['HISTORY'] = '--- Combination of AB spectra ---'
            hdu.header['HISTORY'] = "voffset_pix between A and B {}".format(
                voffset_pix)
        hdu.header.add_history('--- rinput.stored() (BEGIN) ---')
        for item in rinput.stored():
            value = getattr(rinput, item)
            cline = '{}: {}'.format(item, value)
            hdu.header.add_history(cline)
        hdu.header.add_history('--- rinput.stored() (END) ---')
        return result_img

    def set_base_headers(self, hdr):
        newhdr = super(ABBASpectraFastRectwv, self).set_base_headers(hdr)
        # Update EXP to 0
        newhdr['EXP'] = 0
        return newhdr

    def extract_tags_from_obsres(self, obsres, tag_keys):
        # this function is necessary to make use of recipe requirements
        # that are provided as lists
        final_tags = []
        for frame in obsres.frames:
            ref_img = frame.open()
            tag = self.extract_tags_from_ref(
                ref_img,
                tag_keys,
                base=obsres.labels
            )
            final_tags.append(tag)
        return final_tags



#
# Copyright 2010-2014 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# PyEmir is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyEmir is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyEmir.  If not, see <http://www.gnu.org/licenses/>.
#

"""Recipe for the reduction of gain calibration frames."""

import logging
import math

import numpy
import scipy.stats
from astropy.io import fits

from numina.core.requirements import ObservationResultRequirement
from numina.core import Parameter, DataFrame
from numina.exceptions import RecipeError
from numina.core import Result

from emirdrp.core.recipe import EmirRecipe
from emirdrp.instrument.channels import QUADRANTS
from emirdrp.instrument.channels import CHANNELS
from emirdrp.products import MasterGainMap, MasterRONMap

_logger = logging.getLogger('numina.recipes.emir')


class GainRecipe1(EmirRecipe):

    """Detector Gain Recipe.

    Recipe to calibrate the detector gain.
    """
    
    obresult = ObservationResultRequirement()
    region = Parameter('channel', 'Region used to compute: '
                       '(full|quadrant|channel)',
                       choices=['full', 'quadrant', 'channel']
                       )

    gain = Result(MasterGainMap(None, None, None))
    ron = Result(MasterRONMap(None, None))

    def region(self, reqs):
        mm = reqs['region'].tolower()
        if mm == 'full':
            return ((slice(0, 2048), slice(0, 2048)))
        elif mm == 'quadrant':
            return QUADRANTS
        elif mm == 'channel':
            return CHANNELS
        else:
            raise ValueError

    def run(self, rinput):

        resets = []
        ramps = []

        for frame in rinput.obresult.frames:
            if frame.itype == 'RESET':
                resets.append(frame.label)
                _logger.debug('%s is RESET', frame.label)
            elif frame.itype == 'RAMP':
                ramps.append(frame.label)
                _logger.debug('%s is RAMP', frame.label)
            else:
                raise RecipeError('frame is neither a RAMP nor a RESET')

        channels = self.region(rinput)
        result_gain = numpy.zeros((len(channels), ))
        result_ron = numpy.zeros_like(result_gain)

        counts = numpy.zeros((len(ramps), len(channels)))
        variance = numpy.zeros_like(counts)

        last_reset = resets[-1]
        _logger.debug('opening last reset image %s', last_reset)
        last_reset_data = fits.getdata(last_reset)

        for i, di in enumerate(ramps):
            with fits.open(di, mode='readonly') as fd:
                restdata = fd[0].data - last_reset_data
                for j, channel in enumerate(channels):
                    c = restdata[channel].mean()
                    _logger.debug('%f counts in channel', c)
                    counts[i, j] = c
                    v = restdata[channel].var(ddof=1)
                    _logger.debug('%f variance in channel', v)
                    variance[i, j] = v

        for j, _ in enumerate(channels):
            res = scipy.stats.linregress(counts[:, j], variance[:, j])
            slope, intercept, _r_value, _p_value, _std_err = res

            result_gain[j] = 1.0 / slope
            result_ron[j] = math.sqrt(intercept)
        cube = numpy.zeros((2, 2048, 2048))

        for gain, var, channel in zip(result_gain, result_ron, channels):
            cube[0][channel] = gain
            cube[1][channel] = var

        hdu = fits.PrimaryHDU(cube[0])
        hduvar = fits.ImageHDU(cube[1])
        hdulist = fits.HDUList([hdu, hduvar])

        gain = MasterGainMap(mean=result_gain, var=numpy.array([]),
                             frame=DataFrame(hdulist))
        ron = MasterRONMap(mean=result_ron, var=numpy.array([]))
        return self.create_result(gain=gain, ron=ron)



#
# Copyright 2015-2018 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# PyEmir is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyEmir is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyEmir.  If not, see <http://www.gnu.org/licenses/>.
#

"""Bar detection procedures for EMIR"""

import logging
import itertools

import numpy

import emirdrp.instrument.distortions as dist
import scipy.ndimage.measurements as mes
import scipy.ndimage.morphology as morph

from numina.array.utils import expand_region
import numina.array.fwhm as fmod
from numina.array.utils import coor_to_pix_1d
from numina.array.peaks.peakdet import find_peaks_indexes, refine_peaks
from numina.array.utils import slice_create


def find_position(edges, prow, bstart, bend, total=5):
    """Find a EMIR CSU bar position in a edge image.

    Parameters
    ==========
    edges; ndarray,
        a 2d image with 1 where is a border, 0 otherwise
    prow: int,
        reference 'row' of the bars
    bstart: int,
        minimum 'x' position of a bar (0-based)
    bend: int
        maximum 'x' position of a bar (0 based)
    total: int
        number of rows to check near `prow`

    Return
    ======
    list of (x, y) centroids

    """

    nt = total // 2

    # This bar is too near the border
    if prow-nt < 0 or prow + nt >= edges.shape[0]:
        return []

    s2edges = edges[prow-nt:prow+nt+1, bstart:bend]

    structure = morph.generate_binary_structure(2,2) # 8 way conection
    har, num_f = mes.label(s2edges, structure=structure)

    cen_of_mass = mes.center_of_mass(s2edges, labels=har, index=range(1, num_f + 1))

    # center_of_mass returns y, x coordinates

    cen_of_mass_off = [(x + bstart, prow-nt + y) for y,x in cen_of_mass]

    return cen_of_mass_off


def calc_fwhm(img, region, fexpand=3, axis=0):
    """Compute the FWHM in the direction given by axis"""

    # We compute know the FWHM of the slit
    # Given the computed position of the slit
    # Expand 'fexpand' pixels around
    # and cut an slice in the median filtered image

    xpregion = expand_region(region, fexpand, fexpand)
    cslit = img[xpregion]

    # Collapse it
    pslit = cslit.mean(axis=axis)

    # Estimate the background as a flat line
    # starting in pslit[0] and ending in pslit[-1]
    x2 = len(pslit)
    y1, y2 = pslit[0], pslit[-1]
    mslope = (y2-y1) / x2
    # background estimation
    backstim = mslope*numpy.arange(x2) + y1

    # We subtract background
    qslit = pslit-backstim
    # and find the pixel of the maximum
    pidx = numpy.argmax(qslit)
    peak, fwhm = fmod.compute_fwhm_1d_simple(qslit, pidx)
    return fwhm


def simple_prot(x, start):
    """Find the first peak to the right of start"""

    # start must b >= 1

    for i in range(start,len(x)-1):
        a,b,c =  x[i-1], x[i], x[i+1]
        if b - a > 0 and b -c >= 0:
            return i
    else:
        return None


def position_half_h(pslit, cpix, backw=4):
    """Find the position where the value is half of the peak"""

    # Find the first peak to the right of cpix
    next_peak = simple_prot(pslit, cpix)

    if next_peak is None:
        raise ValueError

    dis_peak = next_peak - cpix

    wpos2 = cpix - dis_peak
    wpos1 = wpos2 - backw

    # Compute background in a window of width backw
    # in a position simetrical to the peak
    # around cpix
    left_background = pslit[wpos1:wpos2].min()

    # height of the peak
    height = pslit[next_peak] - left_background

    half_height = left_background + 0.5 * height

    # Position at halg peak, linear interpolation
    vv = pslit[wpos1:next_peak+1] - half_height

    res1, =  numpy.nonzero(numpy.diff(vv > 0))
    i1 = res1[0]

    xint = wpos1 + i1 + (0 - vv[i1]) / (vv[i1+1] - vv[i1])

    return xint, next_peak, wpos1, wpos2, left_background, half_height


def locate_bar_l(icut, epos):
    """Fine position of the left CSU bar"""
    def swap_coor(x):
        return x

    def swap_line(tab):
        return tab

    return _locate_bar_gen(icut, epos,
                           transform1=swap_coor,
                           transform2=swap_line
                           )


def locate_bar_r(icut, epos):
    """Fine position of the right CSU bar"""
    sm = len(icut)

    def swap_coor(x):
        return sm - 1 - x

    def swap_line(tab):
        return tab[::-1]

    return _locate_bar_gen(icut, epos, transform1=swap_coor,
                           transform2=swap_line)


def _locate_bar_gen(icut, epos, transform1, transform2):
    """Generic function for the fine position of the CSU"""

    epos_pix = coor_to_pix_1d(epos)

    # transform ->
    epos_pix_s = transform1(epos_pix)
    icut2 = transform2(icut)
    #

    try:
        res = position_half_h(icut2, epos_pix_s)

        xint_s, next_peak_s, wpos1_s, wpos2_s, background_level, half_height = res
        #

        xint = transform1(xint_s)

        #
        epos_f = xint
        error = 0
    except ValueError:
        error = 2
        epos_f = epos

    return epos_pix, epos_f, error


def char_bar_peak_l(arr_deriv, ypix, bstart, bend, th):
    return _char_bar_peak(arr_deriv, ypix, bstart, bend, th, sign=1)


def char_bar_peak_r(arr_deriv, ypix, bstart, bend, th):
    return _char_bar_peak(arr_deriv, ypix, bstart, bend, th, sign=-1)


def _char_bar_peak(arr_deriv, ypix, bstart, bend, th, sign=1):

    # extract a region to average
    # wy = 3
    # wx = 10
    # Fit the peak with these points
    # wfit = 3

    logger = logging.getLogger(__name__)

    yvpix = numpy.clip(ypix, 0, 2047)

    # Refine at different positions along the slit
    newrefine = []
    wx = 5
    wy = 1
    step = 2 * wy + 1
    # visibility
    #
    intv1 = [0, 2047]
    intv2 = [yvpix - 18, yvpix + 18]
    logger.debug('overlaping interval %s', intv2)

    bar_overlap = overlap(intv1, intv2)
    logger.debug('bar overlaping %.1f', bar_overlap)
    offs = []
    if bar_overlap < 10:
        maxval = 18
    else:
        maxval = 12
    offs2 = range(-step, -maxval, -step)
    offs.extend(reversed(offs2))
    offs.extend(range(0, maxval, step))

    cut = sign * arr_deriv[yvpix, bstart:bend]

    idxs = find_peaks_indexes(cut, window_width=3, threshold=th, fpeak=1)
    logger.debug('found %d peaks over threshold %f', len(idxs), th)

    if len(idxs) == 0:
        logger.debug('no peaks, exit')
        return 0, 0, 0, 0, 0, 1

    # Characterize: use the peak that has the greatest value in the derivative?
    pix_m = cut[idxs].argmax()
    centerx = bstart + idxs[pix_m]
    logger.debug('select the peak with maximum derivative')

    centery = yvpix
    logger.debug('centery is %7.2f at position %7.2f', centery+1,  centerx+1)
    # Refine at the computed center
    xl, fwhm_x, st = refine_bar_centroid(arr_deriv, centerx, centery, wx, wy, th, sign)
    logger.debug('measured values %7.2f (FWHM %7.2f)', xl, fwhm_x)

    if st != 0:
        logger.debug('faillure refining bar centroid, go to next bar')
        # Exiting now, can't refine the centroid
        return centery, centery, xl, xl, fwhm_x, st

    # This is basically to build a list of centers that dont overlap
    for off in offs:
        if 0 <= centery + off <= 2047:
            logger.debug('looping, off %d, measuring at %7.2f', off, centery + off + 1)
            res = refine_bar_centroid(arr_deriv, centerx, centery + off, wx, wy, th, sign)
            logger.debug('looping, measured values %7.2f (FWHM %7.2f)', res[0], res[1])
            newrefine.append(res)
        else:
            logger.debug('looping, off %d, skipping position %7.2f', off, centery + off + 1)

    # this goes in FITS pix coordinates, adding 1
    # filter values with status != 0
    valid_mt = [r[2] == 0 for r in newrefine]
    xcoords_mt = [r[0] + 1 for r in newrefine]
    ycoords_mt = [centery + off + 1 for off in offs]

    ycoords_m = list(itertools.compress(ycoords_mt, valid_mt))
    xcoords_m = list(itertools.compress(xcoords_mt, valid_mt))

    if len(xcoords_m) == 0:
        logger.debug('no valid values to refine')
        return centery, centery, xl, xl, fwhm_x, 3

    logger.debug('transform values from real to virtual')
    xcoords_t, ycoords_t = dist.pvex(xcoords_m, ycoords_m)
    logger.debug('real xcoords are: %s:', xcoords_m)
    logger.debug('real ycoords are: %s:', ycoords_m)
    logger.debug('virtual xcoords are: %s:', xcoords_t)
    logger.debug('virtual ycoords are: %s:', ycoords_t)
    avg_xl_virt = numpy.mean(xcoords_t)
    logger.debug('reference real xcoord is: %s:', xl)
    logger.debug('average virtual xcoord is: %s:', avg_xl_virt)

    centerx_virt, centery_virt = dist.pvex(centerx + 1, centery + 1)

    return centery, centery_virt, xl, avg_xl_virt, fwhm_x, 0


def refine_bar_centroid(arr_deriv, centerx, centery, wx, wy, threshold, sign):
    # Refine values
    logger = logging.getLogger('emir.recipes.bardetect')

    logger.debug('collapsing a %d x %d region', 2 * wx + 1 , 2 * wy + 1)
    #
    slicey = slice_create(centery, wy, start=1, stop=2047)
    slicex = slice_create(centerx, wx, start=1, stop=2047)
    region = arr_deriv[slicey, slicex]

    if region.size == 0:
        logger.debug('region to collapse is empty')
        return 0, 0, 1

    collapsed = sign * region.mean(axis=0)

    # Fine tunning
    idxs_t = find_peaks_indexes(collapsed, window_width=3, threshold=threshold)
    # Use only the peak nearest the original peak
    if len(idxs_t) == 0:
        logger.debug('no peaks after fine-tunning')
        return 0, 0, 2

    dist_t = numpy.abs(idxs_t - wx)
    only_this = dist_t.argmin()
    idxs_p = numpy.atleast_1d(idxs_t[only_this])
    x_t, y_t = refine_peaks(collapsed, idxs_p, window_width=3)

    if len(x_t) == 0:
        logger.debug('no peaks to refine after fitting')
        return 0, 0, 2

    if x_t[0] >= collapsed.shape[0]:
        logger.debug('wrong position %d when refining', x_t[0])
        return 0, 0, 2

    _, fwhm_x = fmod.compute_fwhm_1d_simple(collapsed, x_t[0])

    xl = centerx - wx + x_t[0]
    return xl, fwhm_x, 0


def char_bar_height(arr_deriv_alt, xpos1, xpos2, centery, threshold, wh=35, wfit=3):

    logger = logging.getLogger('emir.recipes.bardetect')
    pcentery = coor_to_pix_1d(centery)
    slicey = slice_create(pcentery, wh, start=1, stop=2047)

    ref_pcentery = pcentery - slicey.start
    mm = arr_deriv_alt[slicey, xpos1:xpos2 + 1].mean(axis=-1)

    idxs_t = find_peaks_indexes(mm, window_width=3,threshold=threshold)
    idxs_u = find_peaks_indexes(-mm, window_width=3,threshold=threshold)
    # Peaks on the right

    status = 0
    npeaks_u = len(idxs_u)
    if npeaks_u == 0:
        # This is a problem, no peak on the right
        b2 = 0
        status = 4
        logger.debug('no bottom border found')
    else:
        # Filter over reference
        g_idxs_u = idxs_u[idxs_u >= ref_pcentery]
        if len(g_idxs_u) == 0:
            logger.debug('no peak over center')
            b2 = 0
            status = 4
        else:
            x_u, y_u = refine_peaks(-mm, g_idxs_u, window_width=wfit)
            # Select the peak with max derivative
            if len(x_u) == 0 or len(y_u) == 0:
                logger.warning("no 1st peak found after refine")
                b2 = 0
                status = 4
            else:
                idmax = y_u.argmax()
                b2 = x_u[idmax]
                b2val = y_u[idmax]
                logger.debug('main border in %f', slicey.start + b2)

    # peaks on the left
    npeaks_t = len(idxs_t)
    if npeaks_t == 0:
        # This is a problem, no peak on the left
        b1 = 0
        logger.debug('no top border found')
        status = 40 + status
    else:
        g_idxs_t = idxs_t[idxs_t <= ref_pcentery]
        if len(g_idxs_t) == 0:
            logger.debug('no peak under center')
            b1 = 0
            status = 40 + status
        else:
            x_t, y_t = refine_peaks(mm, g_idxs_t, window_width=wfit)
            # Select the peak with max derivative
            if len(x_t) == 0 or len(y_t) == 0:
                logger.warning("no 2nd peak found after refine")
                b1 = 0
                status = 40 + status
            else:
                idmax = y_t.argmax()
                b1 = x_t[idmax]
                b1val = y_t[idmax]
                logger.debug('second border in %f', slicey.start + b1)

    return slicey.start + b1, slicey.start + b2, status


def overlap(intv1, intv2):
    """Overlaping of two intervals"""
    return max(0, min(intv1[1], intv2[1]) - max(intv1[0], intv2[0]))


#
# Copyright 2018 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# PyEmir is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyEmir is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyEmir.  If not, see <http://www.gnu.org/licenses/>.
#

"""Wavelength calibration parameters for each grism + filter configuration"""

from __future__ import division, print_function

import numpy as np

from emirdrp.core import EMIR_NAXIS1_ENLARGED
from emirdrp.core import EMIR_VALID_FILTERS
from emirdrp.core import EMIR_VALID_GRISMS


def set_wv_parameters(filter_name, grism_name):
    """Set wavelength calibration parameters for rectified images.

    Parameters
    ----------
    filter_name : str
        Filter name.
    grism_name : str
        Grism name.

    Returns
    -------
    wv_parameters : dictionary
        Python dictionary containing relevant wavelength calibration
        parameters:
        - crpix1_enlarged : float
        - crval1_enlarged : float
        - cdelt1_enlarged : float
        - naxis1_enlarged: int
        - islitlet_min : int
          Minimium slitlet number.
        - islitlet_max : int
          Maximium slitlet number.
        - nbrightlines : python list
          List of integers containing the number of brightlines to be
          used in the initial wavelength calibration.
        - poly_crval1_linear : numpy.polynomial.Polynomial instance
          Polynomial providing the value of CRVAL1_linear as a function
          of csu_bar_slit_center.
        - poly_cdelt1_linear : numpy.polynomial.Polynomial instance
          Polynomial providing the value of CDELT1_linear as a function
          of csu_bar_slit_center.

    """

    # protections
    if filter_name not in EMIR_VALID_FILTERS:
        raise ValueError('Unexpected filter_name:', filter_name)
    if grism_name not in EMIR_VALID_GRISMS:
        raise ValueError('Unexpected grism_name:', grism_name)

    # intialize output
    wv_parameters = {}
    # set parameters
    wv_parameters['crpix1_enlarged'] = 1.0
    if grism_name == "J" and filter_name == "J":
        wv_parameters['islitlet_min'] = 2
        wv_parameters['islitlet_max'] = 54
        wv_parameters['nbrightlines'] = [18]
        wv_parameters['poly_crval1_linear'] = np.polynomial.Polynomial([
            1.25122231e+04,
            -4.83412417e+00,
            5.31657015e-04
        ])
        wv_parameters['poly_cdelt1_linear'] = np.polynomial.Polynomial([
            7.73411692e-01,
            -3.28055653e-05,
            -1.20813896e-08
        ])
        wv_parameters['crval1_enlarged'] = 11200.0000  # Angstroms
        wv_parameters['cdelt1_enlarged'] = 0.77        # Angstroms/pixel
        wv_parameters['naxis1_enlarged'] = EMIR_NAXIS1_ENLARGED  # pixels
        wv_parameters['wvmin_expected'] = None
        wv_parameters['wvmax_expected'] = None
        wv_parameters['wvmin_useful'] = None
        wv_parameters['wvmax_useful'] = None
    elif grism_name == "H" and filter_name == "H":
        wv_parameters['islitlet_min'] = 2
        wv_parameters['islitlet_max'] = 54
        wv_parameters['nbrightlines'] = [12]
        wv_parameters['poly_crval1_linear'] = np.polynomial.Polynomial([
            1.65536274e+04,
            -7.63517173e+00,
            7.74790265e-04
        ])
        wv_parameters['poly_cdelt1_linear'] = np.polynomial.Polynomial([
            1.21327515e+00,
            1.42140078e-05,
            -1.27489119e-07
        ])
        wv_parameters['crval1_enlarged'] = 14500.0000  # Angstroms
        wv_parameters['cdelt1_enlarged'] = 1.2200      # Angstroms/pixel
        wv_parameters['naxis1_enlarged'] = EMIR_NAXIS1_ENLARGED  # pixels
        wv_parameters['wvmin_expected'] = None
        wv_parameters['wvmax_expected'] = None
        wv_parameters['wvmin_useful'] = None
        wv_parameters['wvmax_useful'] = None
    elif grism_name == "K" and filter_name == "Ksp":
        wv_parameters['islitlet_min'] = 2
        wv_parameters['islitlet_max'] = 54
        wv_parameters['nbrightlines'] = [12]
        wv_parameters['poly_crval1_linear'] = np.polynomial.Polynomial([
            2.21044741e+04,
            -1.08737529e+01,
            9.05081653e-04
        ])
        wv_parameters['poly_cdelt1_linear'] = np.polynomial.Polynomial([
            1.72696857e+00,
            2.35009351e-05,
            -1.02164228e-07
        ])
        wv_parameters['crval1_enlarged'] = 19100.0000  # Angstroms
        wv_parameters['cdelt1_enlarged'] = 1.7300      # Angstroms/pixel
        wv_parameters['naxis1_enlarged'] = EMIR_NAXIS1_ENLARGED  # pixels
        wv_parameters['wvmin_expected'] = None
        wv_parameters['wvmax_expected'] = None
        wv_parameters['wvmin_useful'] = None
        wv_parameters['wvmax_useful'] = None
    elif grism_name == "LR" and filter_name == "YJ":
        wv_parameters['islitlet_min'] = 4
        wv_parameters['islitlet_max'] = 55
        wv_parameters['nbrightlines'] = [20]
        wv_parameters['poly_crval1_linear'] = np.polynomial.Polynomial([
            1.04272465e+04,
            -2.33176855e+01,
            6.55101267e-03
        ])
        wv_parameters['poly_cdelt1_linear'] = np.polynomial.Polynomial([
            3.49037727e+00,
            1.26008332e-03,
            -4.66149678e-06
        ])
        wv_parameters['crval1_enlarged'] = 8900.0000   # Angstroms
        wv_parameters['cdelt1_enlarged'] = 3.5600      # Angstroms/pixel
        wv_parameters['naxis1_enlarged'] = 1270        # pixels
        wv_parameters['wvmin_expected'] = 4000
        wv_parameters['wvmax_expected'] = 15000
        wv_parameters['wvmin_useful'] = 8900
        wv_parameters['wvmax_useful'] = 13400
    elif grism_name == "LR" and filter_name == "HK":
        wv_parameters['islitlet_min'] = 4
        wv_parameters['islitlet_max'] = 55
        wv_parameters['nbrightlines'] = [25]
        wv_parameters['poly_crval1_linear'] = np.polynomial.Polynomial([
            2.00704978e+04,
            -4.07702886e+01,
            -5.95247468e-03
        ])
        wv_parameters['poly_cdelt1_linear'] = np.polynomial.Polynomial([
            6.54247758e+00,
            2.09061196e-03,
            -2.48206609e-06
        ])
        wv_parameters['crval1_enlarged'] = 14500.0000  # Angstroms
        wv_parameters['cdelt1_enlarged'] = 6.83        # Angstroms/pixel
        wv_parameters['naxis1_enlarged'] = 1435        # pixels
        wv_parameters['wvmin_expected'] = 8000
        wv_parameters['wvmax_expected'] = 30000
        wv_parameters['wvmin_useful'] = 14500
        wv_parameters['wvmax_useful'] = 24300
    else:
        print("filter_name..:", filter_name)
        print("grism_name...:", grism_name)
        raise ValueError("invalid filter_name and grism_name combination")

    return wv_parameters




import pytest

from numina.user.baserun import run_reduce

# Variable for datamanager_remote
TEST_SET_FILE = "pyemir/pyemir-image-tutorial-v2.tar.gz"


@pytest.mark.result_compare
@pytest.mark.remote_data
@pytest.mark.parametrize("obsid", range(111, 118))
def test_img_tut_block1(datamanager_remote, obsid):
    task = run_reduce(datamanager_remote, obsid)
    return task.result


@pytest.mark.result_compare
@pytest.mark.remote_data
@pytest.mark.parametrize("obsid", range(121, 128))
def test_test_img_tut_block2(datamanager_remote, obsid):
    task = run_reduce(datamanager_remote, obsid)
    return task.result


reduction_reqs1 = {
    'iterations': 1,
    'offsets': 'user_offsets.txt',
    'sky_images': 3
}


reduction_reqs2 = {
    'iterations': 1,
    'sky_images': 3
}

@pytest.mark.result_compare
@pytest.mark.remote_data
@pytest.mark.parametrize("obsid", [20101])
@pytest.mark.parametrize("requirements", [reduction_reqs1, reduction_reqs2])
def test_test_img_tut_block3(datamanager_remote, obsid, requirements):
    task = run_reduce(datamanager_remote, obsid, requirements=requirements)
    return task.result


@pytest.mark.remote_data
def test_test_img_tut_block4(datamanager_remote):
    from numina.util.context import working_directory

    with working_directory(datamanager_remote.basedir):
        with open('control_dump.yaml', 'w') as fp:
            datamanager_remote.backend.dump(fp)
        assert True



#
# Copyright 2008-2019 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

import logging

import numina.ext.gtc
import numina.core.recipes as recipes
import numina.util.flow as flowmod

import emirdrp.core.correctors as cor
import emirdrp.processing.info
import emirdrp.products as prods
import emirdrp.datamodel


class EmirRecipe(recipes.BaseRecipe):
    """Base clase for all EMIR Recipes


    Attributes
    ----------
    logger :
         recipe logger

    datamodel : EmirDataModel

    """
    logger = logging.getLogger(__name__)
    datamodel = emirdrp.datamodel.EmirDataModel()

    def types_getter(self):
        imgtypes = [prods.MasterBadPixelMask,
                    prods.MasterBias,
                    prods.MasterDark,
                    prods.MasterIntensityFlat,
                    prods.MasterSpectralFlat,
                    prods.MasterSky
                    ]
        getters = [cor.get_corrector_p, cor.get_corrector_b, cor.get_corrector_d,
                   [cor.get_corrector_f, cor.get_checker], cor.get_corrector_sf, cor.get_corrector_s]

        return imgtypes, getters

    def get_filters(self):
        import collections
        imgtypes, getters = self.types_getter()
        used_getters = []
        for rtype, getter in zip(imgtypes, getters):
            self.logger.debug('get_filters, %s  %s', rtype, getter)
            if rtype is None:
                # Unconditional
                if isinstance(getter, collections.Iterable):
                    used_getters.extend(getter)
                else:
                    used_getters.append(getter)
            else:
                # Search
                for key, val in self.RecipeInput.stored().items():
                    if isinstance(val.type, rtype):
                        if isinstance(getter, collections.Iterable):
                            used_getters.extend(getter)
                        else:
                            used_getters.append(getter)
                        break
                else:
                    pass
        return used_getters

    def init_filters_generic(self, rinput, getters, ins):
        # with BPM, bias, dark, flat and sky
        if numina.ext.gtc.check_gtc():
            self.logger.debug('running in GTC environment')
        else:
            self.logger.debug('running outside of GTC environment')

        meta = emirdrp.processing.info.gather_info(rinput)
        self.logger.debug('obresult info')
        for entry in meta['obresult']:
            self.logger.debug('frame info is %s', entry)
        correctors = [getter(rinput, meta, ins, self.datamodel) for getter in getters]
        flow = flowmod.SerialFlow(correctors)
        return flow

    def init_filters(self, rinput, ins='EMIR'):
        getters = self.get_filters()
        return self.init_filters_generic(rinput, getters, ins)

    def aggregate_result(self, result, rinput):
        return result



#
# Copyright 2010-2019 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# PyEmir is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyEmir is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyEmir.  If not, see <http://www.gnu.org/licenses/>.
#

"""Recipe for finding cosmetic defects."""

from __future__ import division

import logging
import warnings

import numpy
from astropy.io import fits

from numina import __version__
from numina.core import Parameter
from numina.exceptions import RecipeError
from numina.core import Result
from numina.array.cosmetics import cosmetics, PIXEL_DEAD, PIXEL_HOT
from numina.core.requirements import ObservationResultRequirement
import numina.util.flow
import numina.util.node
import numina.processing as proc

from emirdrp.core.recipe import EmirRecipe
import emirdrp.products as prods
import emirdrp.requirements as reqs


_logger = logging.getLogger(__name__)


def gather_info(hdulist):
    n_ext = len(hdulist)

    # READMODE is STRING
    readmode = hdulist[0].header.get('READMODE', 'undefined')
    bunit = hdulist[0].header.get('BUNIT', 'undefined')
    texp = hdulist[0].header.get('EXPTIME')
    adu_s = True
    if bunit:
        if bunit.lower() == 'adu':
            adu_s = False
        elif bunit.lower() == 'adu/s':
            adu_s = True
        else:
            _logger.warning('Unrecognized value for BUNIT %s', bunit)

    return {'n_ext': n_ext,
            'readmode': readmode,
            'texp': texp,
            'adu_s': adu_s}


class CosmeticsRecipe(EmirRecipe):

    """Detector Cosmetics.

    Recipe to find and tag bad pixels in the detector.
    """

    obresult = ObservationResultRequirement()
    master_bpm = reqs.MasterBadPixelMaskRequirement()
    master_bias = reqs.MasterBiasRequirement()
    master_dark = reqs.MasterDarkRequirement()
    lowercut = Parameter(
        4.0, 'Values below this sigma level are flagged as dead pixels')
    uppercut = Parameter(
        4.0, 'Values above this sigma level are flagged as hot pixels')
    maxiter = Parameter(30, 'Maximum number of iterations')

    ratioframe = Result(prods.ProcessedImage)
    maskframe = Result(prods.MasterBadPixelMask)

    def run(self, rinput):

        # FIXME:
        # We need 2 flats
        # Of different exposure times
        #
        # And their calibrations
        #

        if len(rinput.obresult.frames) < 2:
            raise RecipeError('The recipe requires 2 flat frames')

        iinfo = []
        for frame in rinput.obresult.frames:
            with frame.open() as hdulist:
                iinfo.append(gather_info(hdulist))

        # Loading calibrations
        with rinput.master_bias.open() as hdul:
            readmode = hdul[0].header.get('READMODE', 'undefined')
            if readmode.lower() in ['simple', 'bias']:
                self.logger.debug('loading bias')
                mbias = hdul[0].data
                bias_corrector = proc.BiasCorrector(mbias)
            else:
                self.logger.debug('ignoring bias')
                bias_corrector = numina.util.node.IdNode()

        with rinput.master_dark.open() as mdark_hdul:
            self.logger.debug('loading dark')
            mdark = mdark_hdul[0].data
            dark_corrector = proc.DarkCorrector(mdark)

        flow = numina.util.flow.SerialFlow([bias_corrector, dark_corrector])

        self.logger.info('processing flat #1')
        with rinput.obresult.frames[0].open() as hdul:
            other = flow(hdul)
            f1 = other[0].data.copy() * iinfo[0]['texp'] * 1e-3

        self.logger.info('processing flat #2')
        with rinput.obresult.frames[1].open() as hdul:
            other = flow(hdul)
            f2 = other[0].data.copy() * iinfo[1]['texp'] * 1e-3

        # Preprocess...

        maxiter = rinput.maxiter
        lowercut = rinput.lowercut
        uppercut = rinput.uppercut

        ninvalid = 0
        mask = None

        if mask:
            m = fits.getdata(mask)
            ninvalid = numpy.count_nonzero(m)
        else:
            m = numpy.zeros_like(f1, dtype='int')

        for niter in range(1, maxiter + 1):
            self.logger.debug('iter %d', niter)
            ratio, m, sigma = cosmetics(
                f1, f2, m, lowercut=lowercut, uppercut=uppercut)

            if self.intermediate_results:
                with warnings.catch_warnings():
                    warnings.simplefilter('ignore')
                    fits.writeto('numina-cosmetics-i%02d.fits' %
                                 niter, ratio, overwrite=True)
                    fits.writeto('numina-mask-i%02d.fits' % niter, m, overwrite=True)
                    fits.writeto('numina-sigma-i%02d.fits' %
                                 niter, m * 0.0 + sigma, overwrite=True)
            self.logger.debug(
                'iter %d, invalid points in input mask: %d', niter, ninvalid)
            self.logger.debug('iter %d, estimated sigma is %f', niter, sigma)
            n_ninvalid = numpy.count_nonzero(m)

            # Probably there is something wrong here
            # too much defective pixels
            if ninvalid / m.size >= 0.10:
                # This should set a flag in the output
                msg = 'defective pixels are greater than 10%'
                self.logger.warning(msg)

            if n_ninvalid == ninvalid:
                self.logger.info('convergence reached after %d iterations', niter)
                break
            self.logger.info('new invalid points: %d', n_ninvalid - ninvalid)
            ninvalid = n_ninvalid
        else:
            # This should set a flag in the output
            msg = 'convergence not reached after %d iterations' % maxiter
            self.logger.warning(msg)

        self.logger.info(
            'number of dead pixels %d', numpy.count_nonzero(m == PIXEL_DEAD))
        self.logger.info(
            'number of hot pixels %d', numpy.count_nonzero(m == PIXEL_HOT))

        if self.intermediate_results:
            with warnings.catch_warnings():
                warnings.simplefilter('ignore')
                fits.writeto('numina-cosmetics.fits', ratio, overwrite=True)
                fits.writeto('numina-mask.fits', m, overwrite=True)
                fits.writeto(
                    'numina-sigma.fits', sigma * numpy.ones_like(m), overwrite=True)

        hdu = fits.PrimaryHDU(ratio)
        hdr = hdu.header
        hdr['NUMXVER'] = (__version__, 'Numina package version')
        hdr['NUMRNAM'] = (self.__class__.__name__, 'Numina recipe name')
        hdr['NUMRVER'] = (self.__version__, 'Numina recipe version')
        ratiohdl = fits.HDUList([hdu])

        maskhdu = fits.PrimaryHDU(m)
        hdr = maskhdu.header
        hdr['NUMXVER'] = (__version__, 'Numina package version')
        hdr['NUMRNAM'] = (self.__class__.__name__, 'Numina recipe name')
        hdr['NUMRVER'] = (self.__version__, 'Numina recipe version')
        maskhdl = fits.HDUList([maskhdu])

        res = self.create_result(ratioframe=ratiohdl, maskframe=maskhdl)
        return res



#
# Copyright 2013-2018 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# PyEmir is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyEmir is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyEmir.  If not, see <http://www.gnu.org/licenses/>.
#

"""AIV Recipes for EMIR"""

import logging

from astropy.io import fits
from numina.array.combine import median
from numina.core import Result
from numina.core.requirements import ObservationResultRequirement
from numina.processing.combine import basic_processing_with_combination

from emirdrp.core.recipe import EmirRecipe
import emirdrp.products as prods
import emirdrp.requirements as reqs

_logger = logging.getLogger(__name__)


class SimpleBiasRecipe(EmirRecipe):
    """
    Recipe to process data taken in SimpleBias image Mode.

    Bias images only appear in Simple Readout mode.

    **Outputs:**

     * A combined bias frame, with variance extension.

    **Procedure:**

    The list of images can be readly processed by combining them
    with a median algorithm.
    """

    obresult = ObservationResultRequirement()
    biasframe = Result(prods.MasterBias)

    def run(self, rinput):
        _logger.info('starting simple bias reduction')

        flow = lambda x: x
        hdulist = basic_processing_with_combination(rinput, flow,
                                                    method=median,
                                                    errors=True)

        # update hdu header with
        # reduction keywords
        hdr = hdulist[0].header
        hdr['IMGTYP'] = ('BIAS', 'Image type')
        hdr['NUMTYP'] = ('MASTER_BIAS', 'Data product type')
        hdr['NUMRNAM'] = (self.__class__.__name__, 'Numina recipe name')
        hdr['NUMRVER'] = (self.__version__, 'Numina recipe version')

        _logger.info('simple bias reduction ended')

        # qc is QC.UNKNOWN
        result = self.create_result(biasframe=hdulist)
        return result


class TestBiasCorrectRecipe(EmirRecipe):

    obresult = ObservationResultRequirement()
    master_bpm = reqs.MasterBadPixelMaskRequirement()
    master_bias = reqs.MasterBiasRequirement()
    frame = Result(prods.ProcessedImage)

    def run(self, rinput):
        _logger.info('starting simple bias reduction')

        flow = self.init_filters(rinput)
        hdu = basic_processing_with_combination(rinput, flow, method=median)
        hdr = hdu[0].header
        hdr['NUMRNAM'] = (self.__class__.__name__, 'Numina recipe name')
        hdr['NUMRVER'] = (self.__version__, 'Numina recipe version')
        hdulist = fits.HDUList([hdu])

        result = self.create_result(frame=hdulist)
        return result


class TestRectImageRecipe(EmirRecipe):
    """A Recipe to test GCS handling of rectangular images.

    This appeared as a problem during EMIR first comisioning,
    it was fixed, date 2016-06-21
    """

    obresult = ObservationResultRequirement()
    frame = Result(prods.ProcessedImage)

    def run(self, rinput):
        import numpy
        _logger.info('testing rectangular image')

        data = numpy.zeros((500, 1000), dtype='float32')
        data[200:400, 400:600] = 10000.0
        hdu = fits.PrimaryHDU(data)
        hdulist = fits.HDUList([hdu])
        print("numpy shape of data is", data.shape)

        _logger.info('end testing rectangular image')
        result = self.create_result(frame=hdulist)
        return result


class TestDarkCorrectRecipe(EmirRecipe):

    obresult = ObservationResultRequirement()
    master_bpm = reqs.MasterBadPixelMaskRequirement()
    master_bias = reqs.MasterBiasRequirement()
    master_dark = reqs.MasterDarkRequirement()

    frame = Result(prods.ProcessedImage)

    def run(self, rinput):
        _logger.info('starting simple dark reduction')

        flow = self.init_filters(rinput)
        hdulist = basic_processing_with_combination(rinput, flow,
                                                    method=median)
        hdr = hdulist[0].header
        hdr['NUMRNAM'] = (self.__class__.__name__, 'Numina recipe name')
        hdr['NUMRVER'] = (self.__version__, 'Numina recipe version')

        result = self.create_result(frame=hdulist)

        return result


class TestFlatCorrectRecipe(EmirRecipe):

    obresult = ObservationResultRequirement()
    master_bpm = reqs.MasterBadPixelMaskRequirement()
    master_bias = reqs.MasterBiasRequirement()
    master_dark = reqs.MasterDarkRequirement()
    master_flat = reqs.MasterIntensityFlatFieldRequirement()

    frame = Result(prods.ProcessedImage)

    def run(self, rinput):
        _logger.info('starting simple flat reduction')

        flow = self.init_filters(rinput)
        hdulist = basic_processing_with_combination(rinput, flow, method=median)
        hdr = hdulist[0].header
        self.set_base_headers(hdr)
        result = self.create_result(frame=hdulist)

        return result



#
# Copyright 2008-2018 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

"""Typical requirements of recipes"""

from numina.core import Parameter, Requirement
from numina.core.requirements import ObservationResultRequirement
from numina.types.datatype import ListOfType

import emirdrp.products as prods


# Expose this from numina for convenience
EMIRObservationResultRequirement = ObservationResultRequirement


class EMIRConfigurationRequirement(Requirement):
    """The Recipe requires the configuration of EMIR."""
    def __init__(self):

        super(EMIRConfigurationRequirement,
              self).__init__(prods.EMIRConfigurationType,
                             "EMIR Configuration"
                             )


    def __repr__(self):
        sclass = type(self).__name__
        return "%s(dest=%r, description='%s')" % (sclass,
                                                  self.dest,
                                                  self.description
                                                  )


class MasterBadPixelMaskRequirement(Requirement):
    def __init__(self, optional=True):
        super(MasterBadPixelMaskRequirement,
              self).__init__(prods.MasterBadPixelMask,
                             'Master bad pixel mask',
                             optional=optional
                             )


class MasterBiasRequirement(Requirement):
    def __init__(self, optional=True):
        super(MasterBiasRequirement,
              self).__init__(prods.MasterBias,
                             'Master BIAS image',
                             optional=optional
                             )


class MasterDarkRequirement(Requirement):
    def __init__(self, optional=True):
        super(MasterDarkRequirement,
              self).__init__(prods.MasterDark,
                             'Master DARK image',
                             optional=optional
                             )


class MasterIntensityFlatFieldRequirement(Requirement):
    def __init__(self):
        super(MasterIntensityFlatFieldRequirement,
              self).__init__(prods.MasterIntensityFlat, 'Master intensity flatfield')


class MasterSpectralFlatFieldRequirement(Requirement):
    def __init__(self):
        super(MasterSpectralFlatFieldRequirement,
              self).__init__(prods.MasterSpectralFlat, 'Master spectral flatfield')


class RefinedBoundaryModelParamRequirement(Requirement):
    def __init__(self):
        super(RefinedBoundaryModelParamRequirement,
              self).__init__(prods.RefinedBoundaryModelParam, 'Refined boundary model')


class MasterRectWaveRequirement(Requirement):
    def __init__(self, optional=False):
        super(MasterRectWaveRequirement,
              self).__init__(prods.MasterRectWave, 'Master rectification+wavelength library',
                             optional=optional)


class RectWaveCoeffRequirement(Requirement):
    def __init__(self, optional=False):
        super(RectWaveCoeffRequirement,
              self).__init__(prods.RectWaveCoeff, 'Rectification+wavelength calibration coefficients',
                             optional=optional)


class ListOfRectWaveCoeffRequirement(Requirement):
    def __init__(self, optional=False):
        super(ListOfRectWaveCoeffRequirement,
              self).__init__(ListOfType(prods.RectWaveCoeff), 'List of Rectification+wavelength calibration coefficients',
                             optional=optional)


class MasterSkyRequirement(Requirement):
    def __init__(self, optional=False):
        super(MasterSkyRequirement,
              self).__init__(prods.MasterSky, 'Sky image for subtraction', optional=optional)


class SpectralSkyRequirement(Requirement):
    def __init__(self, optional=False):
        super(SpectralSkyRequirement,
              self).__init__(prods.SkySpectrum, 'Sky spectrum for subtraction', optional=optional)


class Extinction_Requirement(Parameter):
    def __init__(self):
        super(Extinction_Requirement, self).__init__(
            0.0, 'Mean atmospheric extinction'
            )


class SkyImageSepTime_Requirement(Parameter):
    def __init__(self):
        super(SkyImageSepTime_Requirement, self).__init__(
            10.0,
            'Maximum time interval between target and sky images [minutes]'
            )


class Catalog_Requirement(Parameter):
    def __init__(self, optional=True):
        super(Catalog_Requirement, self).__init__(
            [], 'List of x, y coordinates to measure FWHM',
            optional=optional
            )


class Offsets_Requirement(Parameter):
    def __init__(self, optional=True):
        super(Offsets_Requirement, self).__init__(
            [], 'List of pairs of offsets',
            optional=optional
            )




import pytest
import numpy

from emirdrp.processing.bardetect import calc_fwhm
from emirdrp.processing.bardetect import find_position

@pytest.mark.parametrize("axis, result", [(0,7), (1, 16)])
def test_calc_fwhm(axis, result):

    img = numpy.zeros((100, 100))
    img[42:58,45:52] = 100
    region = (slice(40,60), slice(40, 60))
    fwhm = calc_fwhm(img, region, axis=axis)
    assert fwhm == result


def test_find_position_2():
    """Two bars"""

    edges = numpy.zeros((20, 20), dtype='bool')

    edges[3:5,5] = True
    edges[5:8,5] = True

    edges[3:5,12] = True
    edges[5:8,12] = True

    m = find_position(edges, 5, 2, 19, total=5)
    m2 = sorted(m, key=lambda cen: cen[0])

    assert len(m2) == 2

    assert m2[0] == (5.0, 5.0)
    assert m2[1] == (12.0, 5.0)


def test_find_position_0():
    """No edges"""
    edges = numpy.zeros((20, 20), dtype='bool')

    m = find_position(edges, 5, 2, 19, total=5)

    assert len(m) == 0


def test_find_position_1():
    """Only 1 edge."""
    edges = numpy.zeros((20, 20), dtype='bool')

    edges[3:5,5] = True
    edges[5:8,4] = True

    m = find_position(edges, 5, 2, 19, total=5)

    res = (4.4000000000000004, 5.0)

    assert len(m) == 1

    assert m[0] == res



#
# Copyright 2016-2020 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#


"""Twilight Flat Recipe for a list of frames in different filters"""


import uuid
import datetime

import numpy
import astropy.io.fits as fits
from numina.array.combine import median
from numina.array.robustfit import fit_theil_sen
from numina.core import Result
from numina.frame.utils import copy_img
import numina.types.datatype as dt
from numina.processing.combine import basic_processing_with_combination_frames

from emirdrp.processing.info import gather_info_frames
from emirdrp.core.recipe import EmirRecipe
import emirdrp.products as prods
import emirdrp.requirements as reqs


class MultiTwilightFlatRecipe(EmirRecipe):
    """Create a list of twilight flats"""
    obresult = reqs.ObservationResultRequirement()
    master_bpm = reqs.MasterBadPixelMaskRequirement()
    master_bias = reqs.MasterBiasRequirement()
    master_dark = reqs.MasterDarkRequirement()

    twflatframes = Result(dt.ListOfType(prods.MasterIntensityFlat))

    def run(self, rinput):

        results = []
        self.logger.info('starting multiflat flat reduction')

        # Uncomment this line
        # to revert to non-ramp
        # flow = self.init_filters(rinput)
        saturation = 45000.0

        iinfo = gather_info_frames(rinput.obresult.frames)
        image_groups = {}
        self.logger.info('group images by filter')
        for idx, info in enumerate(iinfo):
            filt = info['filter']
            if filt not in image_groups:
                self.logger.debug('new filter %s', filt)
                image_groups[filt] = []
            img = rinput.obresult.frames[idx]
            self.logger.debug('image %s in group %s', img, filt)
            image_groups[filt].append(img)

        for filt, frames in image_groups.items():
            self.logger.info('processing filter %s', filt)

            # Uncomment this line and comment the following
            # to revert to non-ramp
            # res = self.run_per_filter(frames, flow)
            try:
                res = self.run_per_filter_ramp(frames, saturation=saturation)
                results.append(res)
            except ValueError:
                self.logger.info('filter %s cannot be processed', filt)

        self.logger.info('end multiflat flat reduction')
        result = self.create_result(twflatframes=results)

        return result

    def run_per_filter(self, frames, flow):

        errors = True
        self.logger.debug('using errors: %s', errors)
        hdulist = basic_processing_with_combination_frames(frames, flow,
                                                    method=median,
                                                    errors=errors)

        hdr = hdulist[0].header
        self.set_base_headers(hdr)
        mm = hdulist[0].data.mean()
        self.logger.info('mean value of flat is: %f', mm)
        hdr['CCDMEAN'] = mm

        self.logger.debug('normalize image')
        hdulist[0].data /= mm
        if errors:
            self.logger.debug('normalize VAR extension')
            hdulist['variance'].data /= (mm * mm)

        return hdulist

    def run_per_filter_ramp(self, frames, saturation, errors=False):
        imgs = [frame.open() for frame in frames]
        return self.run_img_per_filter_ramp(imgs, saturation, errors)

    def run_img_per_filter_ramp(self, imgs, saturation, errors=False):

        nimages = len(imgs)
        if nimages == 0:
            raise ValueError('len(images) == 0')

        median_frames = numpy.empty((nimages,))
        exptime_frames = []
        utc_frames = []
        # generate 3D cube
        bshape = self.datamodel.shape
        flat_frames = numpy.empty((bshape[0], bshape[1], nimages))
        for idx, image in enumerate(imgs):
            flat_frames[:, :, idx] = image['primary'].data
            exptime_frames.append(image[0].header['EXPTIME'])
            median_frames[idx] = numpy.median(image['primary'].data)
            utc_frames.append(image[0].header['UTC'])

            self.logger.debug(
                'image %d exptime %f median %f UTC %s',
                idx, exptime_frames[idx],
                median_frames[idx],
                utc_frames[-1]
            )

        # filter saturated images
        good_images = median_frames < saturation
        ngood_images = good_images.sum()
        slope_scaled_var = None
        slope_scaled_num = None
        if ngood_images == 0:
            self.logger.warning('We have only %d good images', ngood_images)
            raise ValueError('No images under saturation')
        elif ngood_images < 2:
            self.logger.warning('We have only %d good images', ngood_images)
            # Reference image
            ref_image = imgs[0]
            slope_scaled = numpy.ones(bshape) * exptime_frames[0]

            if errors:
                slope_scaled_var = numpy.zeros_like(slope_scaled)
                slope_scaled_num = numpy.zeros_like(slope_scaled, dtype='int16') + ngood_images
        else:
            nsaturated = nimages - good_images.sum()
            if nsaturated > 0:
                self.logger.debug('we have %d images with median value over saturation (%f)',
                                  nsaturated , saturation)

            m = flat_frames[:,:, good_images]
            # Reshape array to obtain a 2D array
            m_r = m.reshape((bshape[0] * bshape[1], ngood_images))
            self.logger.debug('fitting slopes with Theil-Sen')
            # self.logger.debug('fitting slopes with mean-squares')
            # ll = nppol.polyfit(median_frames[good_images], m_r.T, deg=1)
            ll = self.filter_nsigma(median_frames[good_images], m_r.T)

            slope = ll[1].reshape(bshape)
            base = ll[0].reshape(bshape)

            # First good frame
            index_of_first_good = numpy.nonzero(good_images)[0][0]
            slope_scaled = slope * exptime_frames[index_of_first_good]
            if errors:
                slope_scaled_var = numpy.zeros_like(slope_scaled)
                slope_scaled_num = numpy.zeros_like(slope_scaled, dtype='int16') + ngood_images

        cdata = []
        for idx, img in enumerate(imgs):
            if good_images[idx]:
                cdata.append(img)

        result = self.compose_result(cdata, slope_scaled, errors, slope_scaled_var, slope_scaled_num)

        return result

    def filter_nsigma(self, median_val, image_val, nsigma=10.0, nloop=1):
        # Initial estimation
        ll = fit_theil_sen(median_val, image_val)
        ni = 0
        self.logger.debug('initial estimation')
        while ni < nloop:
            # Prediction
            self.logger.debug('loop %d', ni + 1)
            base, slope = ll
            image_val_pred = base + median_val[:,numpy.newaxis] * slope
            image_diff = image_val - image_val_pred
            # Compute MAD
            mad = compute_mad(image_diff)
            sigma_robust = nsigma * 1.4826 * mad
            self.logger.debug('compute robust std deviation')
            self.logger.debug(
                'min %7.1f max %7.1f mean %7.1f',
                sigma_robust.min(),
                sigma_robust.max(),
                sigma_robust.mean()
            )
            # Check values over sigma
            mask_over = numpy.abs(image_diff) >= sigma_robust[:, numpy.newaxis]
            self.logger.debug('values over sigma: %d', mask_over.sum())
            # Insert expected values in image
            # instead of masking
            image_val[mask_over] = image_val_pred[mask_over]
            #
            self.logger.debug('Theil-Sen fit')
            ll = fit_theil_sen(median_val, image_val)
            ni += 1

        return ll

    def compose_result(self, imgs, slope_scaled, errors=False, slope_scaled_var=None, slope_scaled_num=None):

        self.logger.debug('update result header')
        cnum = len(imgs)
        method_name = 'Theil-Sen'
        result = copy_img(imgs[0])
        base_header = result[0].header
        cdata = imgs

        hdu = result[0]
        hdu.data = slope_scaled
        self.set_base_headers(hdu.header)

        hdu.header['history'] = "Combined %d images using '%s'" % (cnum, method_name)
        hdu.header['history'] = 'Combination time {}'.format(datetime.datetime.utcnow().isoformat())

        for img in cdata:
            hdu.header['history'] = "Image {}".format(self.datamodel.get_imgid(img))

        prevnum = base_header.get('NUM-NCOM', 1)
        hdu.header['NUM-NCOM'] = prevnum * cnum
        hdu.header['UUID'] = str(uuid.uuid1())
        # Headers of last image
        hdu.header['TSUTC2'] = cdata[-1][0].header['TSUTC2']
        # TODO: use BPM to compute mean
        mm = hdu.data.mean()
        self.logger.info('mean value of flat is: %f', mm)
        hdu.header['CCDMEAN'] = mm
        self.logger.debug('normalize image')
        hdu.data /= mm

        if errors:
            varhdu = fits.ImageHDU(slope_scaled_var, name='VARIANCE')
            result.append(varhdu)
            num = fits.ImageHDU(slope_scaled_num, name='MAP')
            result.append(num)
            self.logger.debug('normalize VAR extension')
            varhdu.data /= (mm * mm)
        return result


def compute_mad(x):
    m1 = numpy.median(x, axis=1)
    m2 = numpy.abs(x - m1[:, numpy.newaxis])
    mad = numpy.median(m2, axis=1)
    return mad



#
# Copyright 2018-2020 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# PyEmir is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyEmir is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyEmir.  If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import division
from __future__ import print_function

import argparse
from astropy.io import fits
from datetime import datetime
import logging
import numpy as np
import sys

from numina.array.display.logging_from_debugplot import logging_from_debugplot
from numina.array.wavecalib.apply_integer_offsets import apply_integer_offsets
from numina.array.wavecalib.fix_pix_borders import find_pix_borders
from numina.array.wavecalib.resample import resample_image2d_flux
from numina.frame.utils import copy_img
from numina.tools.arg_file_is_new import arg_file_is_new

import emirdrp.datamodel as datamodel
from emirdrp.instrument.dtu_configuration import DtuConfiguration
from emirdrp.products import RectWaveCoeff

from .set_wv_parameters import set_wv_parameters
from .slitlet2d import Slitlet2D

from numina.array.display.pause_debugplot import DEBUGPLOT_CODES
from emirdrp.core import EMIR_NBARS
from emirdrp.core import EMIR_NPIXPERSLIT_RECTIFIED


def apply_rectwv_coeff(reduced_image,
                       rectwv_coeff,
                       args_resampling=2,
                       args_ignore_dtu_configuration=True,
                       debugplot=0):
    """Compute rectification and wavelength calibration coefficients.

    Parameters
    ----------
    reduced_image : HDUList object
        Image with preliminary basic reduction: bpm, bias, dark and
        flatfield.
    rectwv_coeff : RectWaveCoeff instance
        Rectification and wavelength calibration coefficients for the
        particular CSU configuration.
    args_resampling : int
        1: nearest neighbour, 2: flux preserving interpolation.
    args_ignore_dtu_configuration : bool
        If True, ignore differences in DTU configuration.
    debugplot : int
        Debugging level for messages and plots. For details see
        'numina.array.display.pause_debugplot.py'.

    Returns
    -------
    rectwv_image : HDUList object
        Rectified and wavelength calibrated image.

    """

    logger = logging.getLogger(__name__)

    rectwv_image = copy_img(reduced_image)
    header = rectwv_image[0].header
    image2d = rectwv_image[0].data

    # apply global offsets
    image2d = apply_integer_offsets(
        image2d=image2d,
        offx=rectwv_coeff.global_integer_offset_x_pix,
        offy=rectwv_coeff.global_integer_offset_y_pix
    )

    # check grism and filter
    filter_name = header['filter']
    logger.info('Filter: ' + filter_name)
    if filter_name != rectwv_coeff.tags['filter']:
        raise ValueError('Filter name does not match!')
    grism_name = header['grism']
    logger.info('Grism: ' + grism_name)
    if grism_name != rectwv_coeff.tags['grism']:
        raise ValueError('Grism name does not match!')

    # read the DTU configuration from the image header
    mecs_header = datamodel.get_mecs_header(reduced_image)
    dtu_conf = DtuConfiguration.define_from_header(mecs_header)

    # retrieve DTU configuration from RectWaveCoeff object
    dtu_conf_calib = DtuConfiguration.define_from_dictionary(
        rectwv_coeff.meta_info['dtu_configuration']
    )
    # check that the DTU configuration employed to obtain the calibration
    # corresponds to the DTU configuration in the input FITS file
    if dtu_conf != dtu_conf_calib:
        if args_ignore_dtu_configuration:
            logger.warning('DTU configuration differences found!')
        else:
            logger.warning('DTU configuration from image header:')
            logger.warning(dtu_conf)
            logger.warning('DTU configuration from master calibration:')
            logger.warning(dtu_conf_calib)
            raise ValueError("DTU configurations do not match!")
    else:
        logger.info('DTU configuration match!')

    # valid slitlet numbers
    list_valid_islitlets = list(range(1, EMIR_NBARS + 1))
    for idel in rectwv_coeff.missing_slitlets:
        list_valid_islitlets.remove(idel)
    logger.debug('Valid slitlet numbers:\n' + str(list_valid_islitlets))

    # ---

    # relevant wavelength calibration parameters for rectified and wavelength
    # calibrated image
    wv_parameters = set_wv_parameters(filter_name, grism_name)
    crpix1_enlarged = wv_parameters['crpix1_enlarged']
    crval1_enlarged = wv_parameters['crval1_enlarged']
    cdelt1_enlarged = wv_parameters['cdelt1_enlarged']
    naxis1_enlarged = wv_parameters['naxis1_enlarged']

    # initialize rectified and wavelength calibrated image
    naxis2_enlarged = EMIR_NBARS * EMIR_NPIXPERSLIT_RECTIFIED
    image2d_rectwv = np.zeros((naxis2_enlarged, naxis1_enlarged),
                              dtype='float32')

    # main loop

    logger.info('Applying rectification and wavelength calibration')
    logger.info('RectWaveCoeff uuid={}'.format(rectwv_coeff.uuid))

    cout = '0'
    for islitlet in range(1, EMIR_NBARS + 1):

        if islitlet in list_valid_islitlets:

            # define Slitlet2D object
            slt = Slitlet2D(islitlet=islitlet,
                            rectwv_coeff=rectwv_coeff,
                            debugplot=debugplot)

            # extract (distorted) slitlet from the initial image
            slitlet2d = slt.extract_slitlet2d(image2d)

            # rectify slitlet
            slitlet2d_rect = slt.rectify(slitlet2d, resampling=args_resampling)

            # wavelength calibration of the rectifed slitlet
            slitlet2d_rect_wv = resample_image2d_flux(
                image2d_orig=slitlet2d_rect,
                naxis1=naxis1_enlarged,
                cdelt1=cdelt1_enlarged,
                crval1=crval1_enlarged,
                crpix1=crpix1_enlarged,
                coeff=slt.wpoly
            )

            # minimum and maximum useful row in the full 2d rectified image
            # (starting from 0)
            i1 = slt.iminslt - 1
            i2 = slt.imaxslt

            # minimum and maximum scan in the rectified slitlet
            # (in pixels, from 1 to NAXIS2)
            ii1 = slt.min_row_rectified
            ii2 = slt.max_row_rectified + 1

            # save rectified slitlet in its corresponding location within
            # the full 2d rectified image
            image2d_rectwv[i1:i2, :] = slitlet2d_rect_wv[ii1:ii2, :]

            # include scan range in FITS header
            header['imnslt' + str(islitlet).zfill(2)] = \
                slt.iminslt, 'minimum Y pixel of useful slitlet region'
            header['imxslt' + str(islitlet).zfill(2)] = \
                slt.imaxslt, 'maximum Y pixel of useful slitlet region'

            # determine useful channel region in each spectrum and include
            # that information in FITS header
            jminslt = []
            jmaxslt = []
            for idum in range(ii1, ii2 + 1):
                jminmax = find_pix_borders(
                    slitlet2d_rect_wv[idum, :],
                    sought_value=0
                )
                if jminmax != (-1, naxis1_enlarged):
                    jminslt.append(jminmax[0])
                    jmaxslt.append(jminmax[1])
            if len(jminslt) > 0:
                slt.jminslt = min(jminslt) + 1
                slt.jmaxslt = max(jmaxslt) + 1
            header['jmnslt' + str(islitlet).zfill(2)] = \
                slt.jminslt, 'minimum X pixel of useful slitlet region'
            header['jmxslt' + str(islitlet).zfill(2)] = \
                slt.jmaxslt, 'maximum X pixel of useful slitlet region'

            cout += '.'

        else:

            # include scan and channel range in FITS header
            header['imnslt' + str(islitlet).zfill(2)] = \
                0, 'minimum Y pixel of useful slitlet region'
            header['imxslt' + str(islitlet).zfill(2)] = \
                0, 'maximum Y pixel of useful slitlet region'
            header['jmnslt' + str(islitlet).zfill(2)] = \
                0, 'minimum X pixel of useful slitlet region'
            header['jmxslt' + str(islitlet).zfill(2)] = \
                0, 'maximum X pixel of useful slitlet region'

            cout += 'i'

        if islitlet % 10 == 0:
            if cout != 'i':
                cout = str(islitlet // 10)

        logger.info(cout)

    # update wavelength calibration in FITS header
    logger.info('Updating image header')
    for keyword in ['crval1', 'crpix1', 'crval2', 'crpix2']:
        if keyword in header:
            header.remove(keyword)
    header['crpix1'] = (crpix1_enlarged, 'reference pixel')
    header['crval1'] = (crval1_enlarged, 'central wavelength at crpix1')
    header['cdelt1'] = (cdelt1_enlarged, 'linear dispersion (Angstrom/pixel)')
    header['cunit1'] = ('Angstrom', 'units along axis1')
    header['ctype1'] = 'WAVE'
    header['crpix2'] = (0.0, 'reference pixel')
    header['crval2'] = (0.0, 'central value at crpix2')
    header['cdelt2'] = (1.0, 'increment')
    header['ctype2'] = 'PIXEL'
    header['cunit2'] = ('Pixel', 'units along axis2')
    for keyword in ['cd1_1', 'cd1_2', 'cd2_1', 'cd2_2',
                    'PCD1_1', 'PCD1_2', 'PCD2_1', 'PCD2_2',
                    'PCRPIX1', 'PCRPIX2']:
        if keyword in header:
            header.remove(keyword)

    # update history in FITS header
    header['history'] = 'Boundary parameters uuid:' + \
                        rectwv_coeff.meta_info['origin']['bound_param'][4:]
    if 'master_rectwv' in rectwv_coeff.meta_info['origin']:
        header['history'] = \
            'MasterRectWave uuid:' + \
            rectwv_coeff.meta_info['origin']['master_rectwv'][4:]
    header['history'] = 'RectWaveCoeff uuid:' + rectwv_coeff.uuid
    header['history'] = 'Rectification and wavelength calibration time ' \
                        + datetime.now().isoformat()

    logger.info('Generating rectified and wavelength calibrated image')
    rectwv_image[0].data = image2d_rectwv
    return rectwv_image

    rectwv_image[0].data = image2d_rectwv
    rectwv_image[0].header = header
    return rectwv_image


def main(args=None):
    # parse command-line options
    parser = argparse.ArgumentParser(
        description='description: apply rectification and wavelength '
                    'calibration polynomials for the CSU configuration of a '
                    'particular image'
    )

    # required arguments
    parser.add_argument("fitsfile",
                        help="Input FITS file",
                        type=argparse.FileType('rb'))
    parser.add_argument("--rectwv_coeff", required=True,
                        help="Input JSON file with rectification and "
                             "wavelength calibration coefficients",
                        type=argparse.FileType('rt'))
    parser.add_argument("--outfile", required=True,
                        help="Output FITS file with rectified and "
                             "wavelength calibrated image",
                        type=lambda x: arg_file_is_new(parser, x, mode='wb'))

    # optional arguments
    parser.add_argument("--delta_global_integer_offset_x_pix",
                        help="Delta global integer offset in the X direction "
                             "(default=0)",
                        default=0, type=int)
    parser.add_argument("--delta_global_integer_offset_y_pix",
                        help="Delta global integer offset in the Y direction "
                             "(default=0)",
                        default=0, type=int)
    parser.add_argument("--resampling",
                        help="Resampling method: 1 -> nearest neighbor, "
                             "2 -> linear interpolation (default)",
                        default=2, type=int,
                        choices=(1, 2))
    parser.add_argument("--ignore_dtu_configuration",
                        help="Ignore DTU configurations differences between "
                             "transformation and input image",
                        action="store_true")
    parser.add_argument("--debugplot",
                        help="Integer indicating plotting & debugging options"
                             " (default=0)",
                        default=0, type=int,
                        choices=DEBUGPLOT_CODES)
    parser.add_argument("--echo",
                        help="Display full command line",
                        action="store_true")
    args = parser.parse_args(args)

    if args.echo:
        print('\033[1m\033[31m% ' + ' '.join(sys.argv) + '\033[0m\n')

    # ---

    logging_from_debugplot(args.debugplot)

    # generate RectWaveCoeff object
    rectwv_coeff = RectWaveCoeff._datatype_load(args.rectwv_coeff.name)

    # modify (when requested) global offsets
    rectwv_coeff.global_integer_offset_x_pix += \
        args.delta_global_integer_offset_x_pix
    rectwv_coeff.global_integer_offset_y_pix += \
        args.delta_global_integer_offset_y_pix

    # generate HDUList object
    # read FITS image and its corresponding header
    hdulist = fits.open(args.fitsfile)

    # rectification and wavelength calibration
    reduced_arc = apply_rectwv_coeff(
        hdulist,
        rectwv_coeff,
        args_resampling=args.resampling,
        args_ignore_dtu_configuration=args.ignore_dtu_configuration,
        debugplot=args.debugplot
    )

    # save result
    reduced_arc.writeto(args.outfile, overwrite=True)


if __name__ == "__main__":
    main()



#
# Copyright 2018 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# PyEmir is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyEmir is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyEmir.  If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import division, print_function

import argparse
from datetime import datetime
import logging
import numpy as np
import sys
from uuid import uuid4

from numina.array.display.pause_debugplot import pause_debugplot
from numina.array.display.logging_from_debugplot import logging_from_debugplot
from numina.array.display.polfit_residuals import \
    polfit_residuals_with_sigma_rejection
from numina.array.distortion import ncoef_fmap
from numina.array.stats import summary
from numina.tools.arg_file_is_new import arg_file_is_new
from numina.tools.check_setstate_getstate import check_setstate_getstate

from emirdrp.products import RectWaveCoeff

from numina.array.display.pause_debugplot import DEBUGPLOT_CODES
from emirdrp.core import EMIR_NBARS


def rectwv_coeff_add_longslit_model(rectwv_coeff, geometry, debugplot=0):
    """Compute longslit_model coefficients for RectWaveCoeff object.

    Parameters
    ----------
    rectwv_coeff : RectWaveCoeff instance
        Rectification and wavelength calibration coefficients for a
        particular CSU configuration corresponding to a longslit
        observation.
    geometry : TBD
    debugplot : int
        Debugging level for messages and plots. For details see
        'numina.array.display.pause_debugplot.py'.

    Returns
    -------
    rectwv_coeff : RectWaveCoeff instance
        Updated object with longslit_model coefficients computed.

    """

    logger = logging.getLogger(__name__)

    # check grism and filter
    grism_name = rectwv_coeff.tags['grism']
    logger.info('Grism: ' + grism_name)
    filter_name = rectwv_coeff.tags['filter']
    logger.info('Filter: ' + filter_name)

    # list of slitlets to be computed
    list_valid_islitlets = list(range(1, EMIR_NBARS + 1))
    for idel in rectwv_coeff.missing_slitlets:
        list_valid_islitlets.remove(idel)
    if abs(debugplot) >= 10:
        print('>>> valid slitlet numbers:\n', list_valid_islitlets)

    # ---

    # check that the CSU configuration corresponds to longslit
    csu_bar_slit_center_list = []
    for islitlet in list_valid_islitlets:
        csu_bar_slit_center_list.append(
            rectwv_coeff.contents[islitlet - 1]['csu_bar_slit_center']
        )
    if abs(debugplot) >= 10:
        logger.debug('Checking csu_bar_slit_center values:')
        summary(np.array(csu_bar_slit_center_list), debug=True)
        pause_debugplot(debugplot)

    # ---

    # polynomial coefficients corresponding to the wavelength calibration

    # step 0: determine poldeg_refined, checking that it is the same for
    # all the slitlets
    poldeg_refined_list = []
    for islitlet in list_valid_islitlets:
        poldeg_refined_list.append(
            len(rectwv_coeff.contents[islitlet - 1]['wpoly_coeff']) - 1
        )
    # remove duplicates
    poldeg_refined_list = list(set(poldeg_refined_list))
    if len(poldeg_refined_list) != 1:
        raise ValueError('Unexpected different poldeg_refined found: ' +
                         str(poldeg_refined_list))
    poldeg_refined = poldeg_refined_list[0]

    # step 1: compute variation of each coefficient as a function of
    # y0_reference_middle of each slitlet
    list_poly = []
    for i in range(poldeg_refined + 1):
        xp = []
        yp = []
        for islitlet in list_valid_islitlets:
            tmp_dict = rectwv_coeff.contents[islitlet - 1]
            wpoly_coeff = tmp_dict['wpoly_coeff']
            if wpoly_coeff is not None:
                xp.append(tmp_dict['y0_reference_middle'])
                yp.append(wpoly_coeff[i])
        poly, yres, reject = polfit_residuals_with_sigma_rejection(
            x=np.array(xp),
            y=np.array(yp),
            deg=2,
            times_sigma_reject=5,
            xlabel='y0_rectified',
            ylabel='coeff[' + str(i) + ']',
            title="Fit to refined wavelength calibration coefficients",
            geometry=geometry,
            debugplot=debugplot
        )
        list_poly.append(poly)

    # step 2: use the variation of each polynomial coefficient with
    # y0_reference_middle to infer the expected wavelength calibration
    # polynomial for each rectifified slitlet
    for islitlet in list_valid_islitlets:
        tmp_dict = rectwv_coeff.contents[islitlet - 1]
        y0_reference_middle = tmp_dict['y0_reference_middle']
        list_new_coeff = []
        for i in range(poldeg_refined + 1):
            new_coeff = list_poly[i](y0_reference_middle)
            list_new_coeff.append(new_coeff)
        tmp_dict['wpoly_coeff_longslit_model'] = list_new_coeff

    # ---

    # rectification transformation coefficients aij and bij

    # step 0: determine order_fmap, checking that it is the same for
    # all the slitlets
    order_fmap_list = []
    for islitlet in list_valid_islitlets:
        order_fmap_list.append(
            rectwv_coeff.contents[islitlet - 1]['ttd_order']
        )
    # remove duplicates
    order_fmap_list = list(set(order_fmap_list))
    if len(order_fmap_list) != 1:
        raise ValueError('Unexpected different order_fmap found')
    order_fmap = order_fmap_list[0]

    # step 1: compute variation of each coefficient as a function of
    # y0_reference_middle of each slitlet
    list_poly_ttd_aij = []
    list_poly_ttd_bij = []
    list_poly_tti_aij = []
    list_poly_tti_bij = []
    ncoef_ttd = ncoef_fmap(order_fmap)
    for i in range(ncoef_ttd):
        xp = []
        yp_ttd_aij = []
        yp_ttd_bij = []
        yp_tti_aij = []
        yp_tti_bij = []
        for islitlet in list_valid_islitlets:
            tmp_dict = rectwv_coeff.contents[islitlet - 1]
            ttd_aij = tmp_dict['ttd_aij']
            ttd_bij = tmp_dict['ttd_bij']
            tti_aij = tmp_dict['tti_aij']
            tti_bij = tmp_dict['tti_bij']
            if ttd_aij is not None:
                xp.append(tmp_dict['y0_reference_middle'])
                yp_ttd_aij.append(ttd_aij[i])
                yp_ttd_bij.append(ttd_bij[i])
                yp_tti_aij.append(tti_aij[i])
                yp_tti_bij.append(tti_bij[i])
        poly, yres, reject = polfit_residuals_with_sigma_rejection(
            x=np.array(xp),
            y=np.array(yp_ttd_aij),
            deg=5,
            times_sigma_reject=5,
            xlabel='y0_rectified',
            ylabel='ttd_aij[' + str(i) + ']',
            geometry=geometry,
            debugplot=debugplot
        )
        list_poly_ttd_aij.append(poly)
        poly, yres, reject = polfit_residuals_with_sigma_rejection(
            x=np.array(xp),
            y=np.array(yp_ttd_bij),
            deg=5,
            times_sigma_reject=5,
            xlabel='y0_rectified',
            ylabel='ttd_bij[' + str(i) + ']',
            geometry=geometry,
            debugplot=debugplot
        )
        list_poly_ttd_bij.append(poly)
        poly, yres, reject = polfit_residuals_with_sigma_rejection(
            x=np.array(xp),
            y=np.array(yp_tti_aij),
            deg=5,
            times_sigma_reject=5,
            xlabel='y0_rectified',
            ylabel='tti_aij[' + str(i) + ']',
            geometry=geometry,
            debugplot=debugplot
        )
        list_poly_tti_aij.append(poly)
        poly, yres, reject = polfit_residuals_with_sigma_rejection(
            x=np.array(xp),
            y=np.array(yp_tti_bij),
            deg=5,
            times_sigma_reject=5,
            xlabel='y0_rectified',
            ylabel='tti_bij[' + str(i) + ']',
            geometry=geometry,
            debugplot=debugplot
        )
        list_poly_tti_bij.append(poly)

    # step 2: use the variation of each coefficient with y0_reference_middle
    # to infer the expected rectification transformation for each slitlet
    for islitlet in list_valid_islitlets:
        tmp_dict = rectwv_coeff.contents[islitlet - 1]
        y0_reference_middle = tmp_dict['y0_reference_middle']
        tmp_dict['ttd_order_longslit_model'] = order_fmap
        ttd_aij_longslit_model = []
        ttd_bij_longslit_model = []
        tti_aij_longslit_model = []
        tti_bij_longslit_model = []
        for i in range(ncoef_ttd):
            new_coeff = list_poly_ttd_aij[i](y0_reference_middle)
            ttd_aij_longslit_model.append(new_coeff)
            new_coeff = list_poly_ttd_bij[i](y0_reference_middle)
            ttd_bij_longslit_model.append(new_coeff)
            new_coeff = list_poly_tti_aij[i](y0_reference_middle)
            tti_aij_longslit_model.append(new_coeff)
            new_coeff = list_poly_tti_bij[i](y0_reference_middle)
            tti_bij_longslit_model.append(new_coeff)
        tmp_dict['ttd_aij_longslit_model'] = ttd_aij_longslit_model
        tmp_dict['ttd_bij_longslit_model'] = ttd_bij_longslit_model
        tmp_dict['tti_aij_longslit_model'] = tti_aij_longslit_model
        tmp_dict['tti_bij_longslit_model'] = tti_bij_longslit_model

    # ---

    # update uuid and meta_info in output JSON structure
    rectwv_coeff.uuid = str(uuid4())
    rectwv_coeff.meta_info['creation_date'] = datetime.now().isoformat()

    # return updated object
    return rectwv_coeff


def main(args=None):
    # parse command-line options
    parser = argparse.ArgumentParser()
    # required arguments
    parser.add_argument("--input_rectwv_coeff", required=True,
                        help="Input JSON file with rectification and "
                             "wavelength calibration polynomials "
                             "corresponding to a longslit observation",
                        type=argparse.FileType('rt'))
    parser.add_argument("--output_rectwv_coeff", required=True,
                        help="Output JSON file with updated longslit_model "
                             "coefficients",
                        type=lambda x: arg_file_is_new(parser, x, mode='wt'))

    # optional arguments
    parser.add_argument("--geometry",
                        help="tuple x,y,dx,dy (default 0,0,640,480)",
                        default="0,0,640,480")
    parser.add_argument("--debugplot",
                        help="Integer indicating plotting & debugging options"
                             " (default=0)",
                        default=0, type=int,
                        choices=DEBUGPLOT_CODES)
    parser.add_argument("--echo",
                        help="Display full command line",
                        action="store_true")
    args = parser.parse_args(args)

    if args.echo:
        print('\033[1m\033[31m% ' + ' '.join(sys.argv) + '\033[0m\n')

    # ---

    logging_from_debugplot(args.debugplot)
    logger = logging.getLogger(__name__)

    # geometry
    if args.geometry is None:
        geometry = None
    else:
        tmp_str = args.geometry.split(",")
        x_geom = int(tmp_str[0])
        y_geom = int(tmp_str[1])
        dx_geom = int(tmp_str[2])
        dy_geom = int(tmp_str[3])
        geometry = x_geom, y_geom, dx_geom, dy_geom

    # generate RectWaveCoeff object
    rectwv_coeff = RectWaveCoeff._datatype_load(
        args.input_rectwv_coeff.name)

    # update longslit_model parameters
    rectwv_coeff_updated = rectwv_coeff_add_longslit_model(
        rectwv_coeff=rectwv_coeff,
        geometry=geometry,
        debugplot=args.debugplot
    )

    # save updated RectWaveCoeff object into JSON file
    rectwv_coeff_updated.writeto(args.output_rectwv_coeff.name)
    logger.info('>>> Saving file ' + args.output_rectwv_coeff.name)
    # debugging __getstate__ and __setstate__
    # check_setstate_getstate(rectwv_coeff_updated, args.output_rectwv_coeff.name)


if __name__ == "__main__":
    main()



print("import: 'emirdrp'")
import emirdrp

print("import: 'emirdrp.core'")
import emirdrp.core

print("import: 'emirdrp.instrument'")
import emirdrp.instrument

print("import: 'emirdrp.instrument.configs'")
import emirdrp.instrument.configs

print("import: 'emirdrp.instrument.tests'")
import emirdrp.instrument.tests

print("import: 'emirdrp.processing'")
import emirdrp.processing

print("import: 'emirdrp.processing.tests'")
import emirdrp.processing.tests

print("import: 'emirdrp.processing.wavecal'")
import emirdrp.processing.wavecal

print("import: 'emirdrp.recipes'")
import emirdrp.recipes

print("import: 'emirdrp.recipes.acquisition'")
import emirdrp.recipes.acquisition

print("import: 'emirdrp.recipes.aiv'")
import emirdrp.recipes.aiv

print("import: 'emirdrp.recipes.aiv.tests'")
import emirdrp.recipes.aiv.tests

print("import: 'emirdrp.recipes.auxiliary'")
import emirdrp.recipes.auxiliary

print("import: 'emirdrp.recipes.calibration'")
import emirdrp.recipes.calibration

print("import: 'emirdrp.recipes.engineering'")
import emirdrp.recipes.engineering

print("import: 'emirdrp.recipes.image'")
import emirdrp.recipes.image

print("import: 'emirdrp.recipes.image.tests'")
import emirdrp.recipes.image.tests

print("import: 'emirdrp.recipes.spec'")
import emirdrp.recipes.spec

print("import: 'emirdrp.recipes.spec.tests'")
import emirdrp.recipes.spec.tests

print("import: 'emirdrp.recipes.tests'")
import emirdrp.recipes.tests

print("import: 'emirdrp.simulation'")
import emirdrp.simulation

print("import: 'emirdrp.tests'")
import emirdrp.tests

print("import: 'emirdrp.tools'")
import emirdrp.tools

print("import: 'emirdrp.util'")
import emirdrp.util





import emirdrp.products as prod


def test_tags():
    m = prod.MasterRectWave().query_expr
    assert m.tags() == {'grism', 'filter'}
    m = prod.MasterIntensityFlat().query_expr
    assert m.tags() == {'filter'}


#
# Copyright 2008-2014 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# PyEmir is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyEmir is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyEmir.  If not, see <http://www.gnu.org/licenses/>.
#

import numpy
from astropy import wcs


def offsets_from_wcs(frames, pixref):
    '''Compute offsets between frames using WCS information.

    :parameter frames: sequence of FITS filenames or file descriptors
    :parameter pixref: numpy array used as reference pixel

    The sky world coordinates are computed on *pixref* using
    the WCS of the first frame in the sequence. Then, the
    pixel coordinates of the reference sky world-coordinates
    are computed for the rest of the frames.

    The results is a numpy array with the difference between the
    computed pixel value and the reference pixel. The first line
    of the array is [0, 0], being the offset from the first image
    to itself.

    '''

    result = numpy.zeros((len(frames), pixref.shape[1]))

    with frames[0].open() as hdulist:
        wcsh = wcs.WCS(hdulist[0].header)
        skyref = wcsh.wcs_pix2world(pixref, 1)

    for idx, frame in enumerate(frames[1:]):
        with frame.open() as hdulist:
            wcsh = wcs.WCS(hdulist[0].header)
            pixval = wcsh.wcs_world2pix(skyref, 1)
            result[idx + 1] = -(pixval[0] - pixref[0])

    return result


def offsets_from_wcs_imgs(imgs, pixref):

    result = numpy.zeros((len(imgs), pixref.shape[1]))

    ref = imgs[0]
    wcsh = wcs.WCS(ref[0].header)
    skyref = wcsh.wcs_pix2world(pixref, 1)

    for idx, img in enumerate(imgs[1:]):
        wcsh = wcs.WCS(img[0].header)
        pixval = wcsh.wcs_world2pix(skyref, 1)
        result[idx + 1] = -(pixval[0] - pixref[0])

    return result


def reference_pix_from_wcs(frames, pixref, origin=1):
    """Compute reference pixels between frames using WCS information.

    The sky world coordinates are computed on *pixref* using
    the WCS of the first frame in the sequence. Then, the
    pixel coordinates of the reference sky world-coordinates
    are computed for the rest of the frames.

    The results is a list with the position of the reference pixel
    in each image

    """

    result = []

    with frames[0].open() as hdulist:
        wcsh = wcs.WCS(hdulist[0].header)
        skyref = wcsh.wcs_pix2world([pixref], origin)
        result.append(pixref)

    for idx, frame in enumerate(frames[1:]):
        with frame.open() as hdulist:
            wcsh = wcs.WCS(hdulist[0].header)
            pixval = wcsh.wcs_world2pix(skyref, origin)
            result.append(tuple(pixval[0]))

    return result


def reference_pix_from_wcs_imgs(imgs, pixref, origin=1):
    """Compute reference pixels between frames using WCS information.

    The sky world coordinates are computed on *pixref* using
    the WCS of the first frame in the sequence. Then, the
    pixel coordinates of the reference sky world-coordinates
    are computed for the rest of the frames.

    The results is a list with the position of the reference pixel
    in each image

    """

    result = []

    refimg = imgs[0]

    wcsh = wcs.WCS(refimg[0].header)
    skyref = wcsh.wcs_pix2world([pixref], origin)
    result.append(pixref)

    for idx, img in enumerate(imgs[1:]):
        wcsh = wcs.WCS(img[0].header)
        pixval = wcsh.wcs_world2pix(skyref, origin)
        result.append(tuple(pixval[0]))

    return result



#
# Copyright 2019 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# PyEmir is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PyEmir is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyEmir.  If not, see <http://www.gnu.org/licenses/>.
#

from __future__ import division
from __future__ import print_function

import logging
import numpy as np
from scipy.stats import norm

from numina.array.distortion import fmap

from emirdrp.core import EMIR_NAXIS1
from emirdrp.core import EMIR_NAXIS2
from emirdrp.core import EMIR_NBARS


def synthetic_lines_rawdata(catlines_all_wave,
                            catlines_all_flux,
                            list_useful_slitlets,
                            rectwv_coeff,
                            ycorrection_pix=2.0,
                            sigma_gauss_pix = 1.5):
    """Generate synthetic MOS data with arc/OH lines.

    Parameters
    ----------
    catlines_all_wave : numpy array
        Array with wavelengths of the lines to be overplotted.
    catlines_all_flux : numpy array
        Array with fluxes (arbitrary units) of the lines to be
        overplotted.
    list_useful_slitlets : list of integers
        List with numbers of valid slitlets.
    rectwv_coeff : RectWaveCoeff instance
        Rectification and wavelength calibration coefficients for the
        particular CSU configuration.
    ycorrection_pix : float
        Approximate number of pixels from frontier to boundary.
    sigma_gauss_pix : float
        Sigma of the Gaussians to be employed to reproduce the
        emission lines.

    Returns
    -------
    simulated_image : numpy array
        Simulated image with arc/OH lines.

    """

    logger = logging.getLogger(__name__)
    logger.info('generating synthetic MOS image')

    simulated_image = np.zeros((EMIR_NAXIS2, EMIR_NAXIS1))

    global_integer_offset_x_pix = rectwv_coeff.global_integer_offset_x_pix
    global_integer_offset_y_pix = rectwv_coeff.global_integer_offset_y_pix

    nside_pix = int(sigma_gauss_pix * 5.0 + 0.5)

    cout = '0'
    for islitlet in range(1, EMIR_NBARS + 1):
        if islitlet in list_useful_slitlets:
            cout += '.'
            dumdict = rectwv_coeff.contents[islitlet - 1]
            # check pseudo-longslit with previous slitlet
            center = dumdict['csu_bar_slit_center']
            width = dumdict['csu_bar_slit_width']
            connected_prev = False
            if islitlet > 1:
                dumdict_prev = rectwv_coeff.contents[islitlet - 2]
                center_prev = dumdict_prev['csu_bar_slit_center']
                width_prev = dumdict_prev['csu_bar_slit_width']
                if abs(center - center_prev) < width/2:
                    if abs(width - width_prev) < width/2:
                        connected_prev = True
            # check pseudo-longslit with next slitlet
            connected_next = False
            if islitlet < EMIR_NBARS:
                dumdict_next = rectwv_coeff.contents[islitlet]
                center_next = dumdict_next['csu_bar_slit_center']
                width_next = dumdict_next['csu_bar_slit_width']
                if abs(center - center_next) < width/2:
                    if abs(width - width_next) < width/2:
                        connected_next = True
            # wavelength calibration parameters
            crval1_linear = dumdict['crval1_linear']
            cdelt1_linear = dumdict['cdelt1_linear']
            crvaln_linear = crval1_linear + (EMIR_NAXIS1 - 1) * cdelt1_linear
            # rectification parameters
            bb_ns1_orig = dumdict['bb_ns1_orig']
            ttd_order = dumdict['ttd_order']
            aij = dumdict['ttd_aij']
            bij = dumdict['ttd_bij']
            min_row_rectified = float(dumdict['min_row_rectified'])
            max_row_rectified = float(dumdict['max_row_rectified'])
            wpoly_coeff = dumdict['wpoly_coeff']
            # store coordinates for each line
            x1 = []
            y1 = []
            x2 = []
            y2 = []
            flist = []
            for wave, flux in zip(catlines_all_wave, catlines_all_flux):
                if crval1_linear <= wave <= crvaln_linear:
                    tmp_coeff = np.copy(wpoly_coeff)
                    tmp_coeff[0] -= wave
                    tmp_xroots = np.polynomial.Polynomial(tmp_coeff).roots()
                    for dum in tmp_xroots:
                        if np.isreal(dum):
                            dum = dum.real
                            if 1 <= dum <= EMIR_NAXIS1:
                                x1.append(dum)
                                y1.append(min_row_rectified)
                                x2.append(dum)
                                y2.append(max_row_rectified)
                                flist.append(flux)
            xx1, yy1 = fmap(ttd_order, aij, bij, np.array(x1), np.array(y1))
            xx1 -= global_integer_offset_x_pix
            yy1 += bb_ns1_orig - global_integer_offset_y_pix
            if not connected_prev:
                yy1 += ycorrection_pix
            xx2, yy2 = fmap(ttd_order, aij, bij, np.array(x2), np.array(y2))
            xx2 -= global_integer_offset_x_pix
            yy2 += bb_ns1_orig - global_integer_offset_y_pix
            if not connected_next:
                yy2 -= ycorrection_pix

            for xx1_, xx2_, yy1_, yy2_, flux in zip(xx1, xx2, yy1, yy2, flist):
                slope = (xx2_ - xx1_) / (yy2_ - yy1_)
                iyy1_ = int(yy1_ + 0.5)
                if yy1_ <= float(iyy1_):
                    fracpix1 = (float(iyy1_) - yy1_)
                    iyy1_ -= 1
                else:
                    fracpix1 = 1.0 - (yy1_ - float(iyy1_))
                iyy2_ = int(yy2_ + 0.5)
                if yy2_ <= float(iyy2_):
                    fracpix2 = 1.0 - (float(iyy2_) - yy2_)
                else:
                    fracpix2 = (yy2_ - float(iyy2_))
                    iyy2_ += 1
                for iyy_ in range(iyy1_, iyy2_ + 1):
                    if iyy_ == iyy1_:
                        fracpix = fracpix1
                    elif iyy_ == iyy2_:
                        fracpix = fracpix2
                    else:
                        fracpix = 1.0
                    icenter = iyy_ - 1
                    if 0 <= icenter <= EMIR_NAXIS2 - 1:
                        xx0_ = xx1_ + slope * (iyy_ - yy1_)
                        xcenter = int(xx0_ + 0.5)
                        npix = 2 * nside_pix + 1
                        xpix = xcenter + np.arange(-nside_pix, nside_pix + 1)
                        left_border = xpix - xx0_ - 0.5
                        right_border = left_border + 1.0
                        borders = np.concatenate((left_border,
                                                  right_border))
                        area = norm.cdf(borders, scale=sigma_gauss_pix)
                        for i in range(npix):
                            xdum = xpix[i]
                            if 1 <= xdum <= EMIR_NAXIS1:
                                simulated_image[icenter, xdum - 1] += \
                                    fracpix * flux * (area[i + npix] - area[i])
        else:
            cout += 'i'

        if islitlet % 10 == 0:
            if cout != 'i':
                cout = str(islitlet // 10)

        logger.info(cout)

    return simulated_image




import pkgutil

import pytest

from ..csuconf import CSUConf
from ..csuconf import CSUBarModel, CSUBarModelL, CSUBarModelR
from ..csuconf import PhysicalBar, PhysicalBarL, PhysicalBarR
from ..csuconf import LogicalSlit, TargetType
from ..csuconf import create_bar_models, read_csu_from_header
from ..csuconf import merge_slits
from ..csuconf import EMIR_NBARS


import numpy
try:
    import StringIO as S
except ImportError:
    import io as S


def create_test_header0():
    hdr = {}
    for i in range(55):
        hdr["CSUP{}".format(i + 1)] = -100
    for i in range(55, 110):
        hdr["CSUP{}".format(i + 1)] = -100
    return hdr


def create_test_header1():
    hdr = create_test_header0()

    hdr["SLIFL12"] = 2
    hdr["SLIFL13"] = 2
    hdr["SLIFL33"] = 2
    hdr["SLIFL44"] = 2
    hdr["SLIFL45"] = 2

    hdr["XRSLI12"] = 200.1
    hdr["XRSLI13"] = 200.1
    hdr["YRSLI12"] = 300.1
    hdr["YRSLI13"] = 300.1

    hdr["XRSLI33"] = 1300.1
    hdr["YRSLI33"] = 1300.1
    hdr["XRSLI44"] = 1900.1
    hdr["YRSLI44"] = 1850.1
    hdr["XRSLI45"] = 1900.1
    hdr["YRSLI45"] = 1850.1
    #
    hdr["XVSLI12"] = 200.1
    hdr["XVSLI13"] = 200.1
    hdr["YVSLI12"] = 300.1
    hdr["YVSLI13"] = 300.1
    #
    hdr["XRSLI33"] = 1300.1
    hdr["YRSLI33"] = 1300.1
    hdr["XRSLI44"] = 1900.1
    hdr["YRSLI44"] = 1850.1
    hdr["XRSLI45"] = 1900.1
    hdr["YRSLI45"] = 1850.1
    return hdr


@pytest.mark.parametrize("hdr, nslits",[(create_test_header0(), 55), (create_test_header1(), 53)])
def test_csubar(hdr, nslits):
    dumdata = pkgutil.get_data('emirdrp.instrument.configs', 'bars_nominal_positions_test.txt')
    ss = S.StringIO(dumdata.decode('utf8'))
    bars_nominal_positions = numpy.loadtxt(ss)
    barmodel = create_bar_models(bars_nominal_positions)
    csu_conf = read_csu_from_header(barmodel, hdr)
    assert len(csu_conf.slits) == nslits


@pytest.mark.parametrize("hdr, nslits",[(create_test_header0(), 55), (create_test_header1(), 53)])
def test_merge_bars(hdr, nslits):
    mm = []
    for idx in range(1, EMIR_NBARS + 1):

        # References from header
        try:
            slit_t = hdr["SLIFL%d" % idx]
            target_type = TargetType(slit_t)
        except KeyError:
            target_type = TargetType.UNKNOWN

        xref = hdr.get("XRSLI%d" % idx, -100) - 1
        yref = hdr.get("YRSLI%d" % idx, -100) - 1
        target_coordinates = (xref, yref)

        xref = hdr.get("XVSLI%d" % idx, -100) - 1
        yref = hdr.get("YVSLI%d" % idx, -100) - 1
        target_coordinates_v = (xref, yref)

        mm.append((idx, target_type, target_coordinates, target_coordinates_v))

    bag = merge_slits(mm, max_slits=3, tol=1e-2)
    assert len(bag) == nslits


def test_csuconf1():
    dumdata = pkgutil.get_data('emirdrp.instrument.configs', 'bars_nominal_positions_test.txt')
    ss = S.StringIO(dumdata.decode('utf8'))
    bars_nominal_positions = numpy.loadtxt(ss)
    hdr = create_test_header1()
    barmodel = create_bar_models(bars_nominal_positions)
    csu_conf = read_csu_from_header(barmodel, hdr)
    assert csu_conf.is_open()



#
# Copyright 2011-2020 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#


"""
Extra functions
"""

import logging
import contextlib

import numpy
import astropy.io.fits as fits

_logger = logging.getLogger(__name__)


def generate_bpm_hdu(hdu):
    """Generate a BPM extension copying an existing HDU"""
    return generate_bpm_data(hdu.data, dtype='uint8', header=hdu.header)


def generate_empty_bpm_hdu(hdu, dtype='uint8'):
    """Generate a BPM extension filled with zeros"""
    data = numpy.zeros_like(hdu.data, dtype)
    return generate_bpm_data(data)


def generate_bpm_data(data, dtype='uint8', header=None):
    """Generate a BPM extension from data."""
    # Mask is assumed to be UINT8
    # casting here
    data = data.astype(dtype)
    hdu_bpm = fits.ImageHDU(
        data,
        header=header
    )

    hdu_bpm.header['EXTNAME'] = 'BPM'
    return hdu_bpm



#
# Copyright 2016-2020 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

EMIR_BIAS_MODES = ['simple', 'bias', 'single']
EMIR_READ_MODES = ['simple', 'bias', 'single', 'cds', 'fowler', 'ramp']

EMIR_PLATESCALE = 0.1944 # arcsec/pixel

# plate scale, focal plane NASMYTH_A, arcsec / mm
# keyword PSCFP
NASMYTH_A_PS = 1.212 ## arcsec / mm
EMIR_CSU_PS = 1.1746 ## arcsec / mm

# plate scale, detector, arcsec / mm
EMIR_DETECTOR_PS = 10.8
# plate scale, detector, arcsec / pix
EMIR_PS_PIX = EMIR_PLATESCALE
# Pixel size, mm
# PXSZDET
EMIR_PIXSCALE = 18.0 # um
EMIR_PIXSIZE_DETECTOR = EMIR_PIXSCALE * 1e-3

# Pixel projected in FP, mm
# keyword PXSZCSU
EMIR_PIXSIZE_CSU = EMIR_PLATESCALE / EMIR_CSU_PS

# plate scale, detector, arcsec / pix
EMIR_DET_PS_PIX = EMIR_DETECTOR_PS * EMIR_PIXSIZE_DETECTOR

EMIR_GAIN = 5.0 # ADU / e-
EMIR_RON = 5.69 # ADU

EMIR_NBARS = 55
EMIR_NAXIS1 = 2048
EMIR_NAXIS1_ENLARGED = 3400
EMIR_NAXIS2 = 2048

EMIR_NPIXPERSLIT_RECTIFIED = 38
EMIR_MINIMUM_SLITLET_WIDTH_MM = 0.0
EMIR_MAXIMUM_SLITLET_WIDTH_MM = 1000.0
EMIR_VALID_GRISMS = ['J', 'H', 'K', 'LR']
EMIR_VALID_FILTERS = ['J', 'H', 'Ksp', 'YJ', 'HK']




#
# Copyright 2008-2020 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

"""The EMIR Data Reduction Pipeline"""

import logging

__version__ = '0.15.0'


# Top level NullHandler
logging.getLogger("emirdrp").addHandler(logging.NullHandler())



#
# Copyright 2016-2020 Universidad Complutense de Madrid
#
# This file is part of PyEmir
#
# SPDX-License-Identifier: GPL-3.0+
# License-Filename: LICENSE.txt
#

"""Datamodel for EMIR and related functions"""

from __future__ import division

import logging

import numina.datamodel as dm
import astropy.wcs

import emirdrp.instrument


_logger = logging.getLogger(__name__)


class EmirDataModel(dm.DataModel):
    """Data model of EMIR."""

    meta_dinfo_headers = [
        'readmode',
        'exptime',
        'grism',
        'filter',
        'mode',
        'tstamp',
        'uuid1',
        'uuid2',
        'skyadd'
    ]

    query_attrs = {
        'filter': dm.QueryAttribute('filter', str),
        'grism': dm.QueryAttribute('grism', str),
        'insconf': dm.QueryAttribute('insconf', str)
    }


    def __init__(self):
        defaults = self.default_mappings()
        # FIXME: this should be computed
        defaults['darktime'] = 'exptime'

        instrument_mappings = {
            'readmode': ('READMODE', 'undefined'),
            'grism': ('GRISM', 'undefined'),
            'filter': ('FILTER', 'undefined'),
            'tstamp': ('TSTAMP', 'undefined'),
            'uuid1': ('UUID', 'undefined'),
            'uuid2': ('EMIRUUID', 'undefined'),
            'skyadd': ('SKYADD', True),
            'insconf': ('INSCONF', None),
            'blckuuid': lambda img: None,
        }

        defaults.update(instrument_mappings)

        super(EmirDataModel, self).__init__(
            'EMIR',
            defaults
        )

    @property
    def shape(self):
        return (2048, 2048)

    def do_sky_correction(self, img):
        header = img['primary'].header
        return header.get('SKYADD', True)

    def add_computation_time(self, img, time1, time2):
        img[0].header['NUMUTC1'] = time1.isoformat()
        img[0].header['NUMUTC2'] = time2.isoformat()
        return img


def get_mecs_header(hdulist):
    if 'MECS' in hdulist:
        return hdulist['MECS'].header
    else:
        return hdulist[0].header


def get_dtur_from_img(hdulist):
    mecs_header = get_mecs_header(hdulist)
    return get_dtur_from_header(mecs_header)


def get_dtur_from_header(hdr):
    from emirdrp.instrument.dtuconf import DtuConf
    dtuconf = DtuConf.from_header(hdr)
    return dtuconf.coor, dtuconf.coor_r


def get_csup_from_header(hdr):
    # CSUP keys
    default = 0.0
    first_idx = 1
    last_idx = 110
    _logger.info('getting CSUP keys from header')
    values = []
    for idx in range(first_idx, last_idx+1):
        key = "CSUP{}".format(idx)
        values.append(hdr.get(key, default))

    return values


def get_cs_from_header(hdr):
    # CS keys
    default = 0.0
    first_idx = 0
    last_idx = 439
    _logger.info('getting CS keys from header')
    values = []
    for idx in range(first_idx, last_idx+1):
        key = "CS_{}".format(idx)
        values.append(hdr.get(key, default))

    return values
